// play_TLM2_membufP.cpp: conceptual audio/video stream decoding example
// Author: RISC project members, CECS
//
// 03/15/21 RD	version with polling instead of events (interrupts)
// 03/14/21 RD	option for memory and device access delay
// 03/12/21 RD	version with FIFO buffers in memory, sync'd by events
// 02/04/19 RD  version with 1 memory, hierarchical socket binding
// 02/03/19 RD  added validation of display and speaker data
// 01/29/19 RD  version with 3 memories, cross-hierarchy socket binding
// 01/28/18 RD  improved generic payload to meet Doulos recommendations
//              and applied some cleaning for better presentation
// 27/11/18 ZC  use sockets to replace ports and channels
// 06/05/18 RD  confirmed clean sc_time constants with GCC 6.3.1
// 04/27/18 RD  clean sc_time constants for newer compilers (e.g. GCC 4.9.2)
// 04/26/18 RD  merged tracing version into main file (define TRACE_FILE)
// 04/24/18 DM  added tracing of stream data (sc_trace is now supported)
// 07/10/17 RD  split the delay into two parts (as suggested by ZC)
// 04/05/17 RD  removed sc_stop from stimulus, let monitor terminate;
//              added COUNT_FRAMES to ensure proper completion
// 03/08/17 RD  avoid duplicated DELAY macros, set BUFSIZE to 10
// 01/13/17 GL  separate version play_fifo.cpp with sc_fifo channels
// 10/04/16 RD  removed obsolete C artifacts from the user-defined queue channel
// 09/30/16 RD  cleanup for release 0.3.0
// 03/07/15 RD  bug fix: use sc_stop() to quit cleanly (exit() is not MT-safe)
// 10/30/15 RD  polished for release 0.2.1
// 09/04/15 GL  use 'long long' (not 'double') to avoid floating-point overflow
//              (FPU exception handling shows high performance variation;
//              moreover, integer arithmetic is more realistic for decoder)
// 08/30/15 GL  initial version based on SpecC example play.sc by RD

#include <stdio.h>
#include "systemc.h"
#include "tlm.h"                               // TLM headers
#include "tlm_utils/simple_initiator_socket.h"
#include "tlm_utils/simple_target_socket.h"
#include <string>
using namespace sc_core;
using namespace std;

// frame rates
#define AUDIO_DELAY      26120, SC_US
#define VIDEO_DELAY      33330, SC_US
#define HALF_AUDIO_DELAY 26120/2, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US
#define POLLING_DELAY       50, SC_US

// simulated time
#define CLIP_LENGTH     10, SC_SEC

// simulated workload unit
#define AUDIOLOAD        7500000
#define VIDEOLOAD       10000000

// memory and device access delays
#define MEM_READ_ACCESS_DELAY   0, SC_NS
#define MEM_WRITE_ACCESS_DELAY  0, SC_NS
#define DEV_WRITE_ACCESS_DELAY  0, SC_NS

// illustrate and check frame decoding
//#define PRINT_TIME
//#define PRINT_FRAMES
//#define COUNT_FRAMES
//#define TRACE_FILE
//#define DUMP_FRAMES
//#define CHECK_FRAMES

#ifdef DUMP_FRAMES
FILE *fh[3];
#endif
#ifdef CHECK_FRAMES
#  include "v.frames"
#  include "l.frames"
#  include "r.frames"
#endif

#define DEBUG_PRINTF(s,...)
//#define DEBUG_PRINTF(s,...) printf("%9llu.%4u:%s: " s "\n", \
	sc_time_stamp().value()/1000000, sc_delta_count(), name(), __VA_ARGS__)

// memory sizes and segmentation
#define MEMORY_SIZE     0x10000

#define ADDRESS_VIDEO   0x00000
#define ADDRESS_VIDEO_SENT_CNT (ADDRESS_VIDEO)
#define ADDRESS_VIDEO_RCVD_CNT (ADDRESS_VIDEO_SENT_CNT + sizeof(unsigned))
#define ADDRESS_VIDEO_BUFFER   (ADDRESS_VIDEO_RCVD_CNT + sizeof(unsigned))
#define ADDRESS_VIDEO_FRAME(n) (ADDRESS_VIDEO_BUFFER + (n)*sizeof(long long))
#define BUFFER_SIZE_VIDEO 3

#define BASE_ADDRESS_AUDIO_L  0x00100
#define BASE_ADDRESS_AUDIO_R  0x00200
#define OFFSET_AUDIO_SENT_CNT 0
#define OFFSET_AUDIO_RCVD_CNT (OFFSET_AUDIO_SENT_CNT + sizeof(unsigned))
#define OFFSET_AUDIO_BUFFER   (OFFSET_AUDIO_RCVD_CNT + sizeof(unsigned))
#define OFFSET_AUDIO_FRAME(n) (OFFSET_AUDIO_BUFFER + (n)*sizeof(long long))
#define BUFFER_SIZE_AUDIO 5

SC_MODULE(Memory4p)
{
  tlm_utils::simple_target_socket<Memory4p>  s0;
  tlm_utils::simple_target_socket<Memory4p>  s1;
  tlm_utils::simple_target_socket<Memory4p>  s2;
  tlm_utils::simple_target_socket<Memory4p>  s3;

//....TODO: #pragma RISC M conflict-free
  unsigned char *M;

  SC_CTOR(Memory4p)
  {
    M = new unsigned char[MEMORY_SIZE];
    s0.register_b_transport(this, &Memory4p::access);
    s1.register_b_transport(this, &Memory4p::access);
    s2.register_b_transport(this, &Memory4p::access);
    s3.register_b_transport(this, &Memory4p::access);
  }

  void access(tlm::tlm_generic_payload& trans, sc_time& delay)
  {
    tlm::tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    unsigned char   *ptr = trans.get_data_ptr();
    unsigned int     len = trans.get_data_length();
    unsigned int     wid = trans.get_streaming_width();
    unsigned char   *byt = trans.get_byte_enable_ptr();
    if (  (adr + len <= MEMORY_SIZE)
          &&(ptr != 0)
          &&(wid == len)
          &&(byt == 0))
    {
      if (cmd == tlm::TLM_READ_COMMAND)
      {
        memcpy(ptr, &M[adr], len);
        trans.set_response_status(tlm::TLM_OK_RESPONSE);
        delay += sc_time(MEM_READ_ACCESS_DELAY);
      }
      else if (cmd == tlm::TLM_WRITE_COMMAND)
      {
        memcpy(&M[adr], ptr, len);
        trans.set_response_status(tlm::TLM_OK_RESPONSE);
        delay += sc_time(MEM_WRITE_ACCESS_DELAY);
      }
      else
      {
        trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
      }
    }
    else
    {
      trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
    }
  }
};

SC_MODULE(Stimulus)
{
  tlm_utils::simple_initiator_socket<Stimulus> MultiMediaStream;

  long long    video;
  unsigned int VframesSent, VframesRcvd;
  long long    audioLeft;
  unsigned int LframesSent, LframesRcvd;
  long long    audioRight;
  unsigned int RframesSent, RframesRcvd;

  void MemAccess(
	tlm::tlm_command cmd,
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	void *data,
	unsigned int len)
  {
    sc_time delay = SC_ZERO_TIME;
    tlm::tlm_generic_payload trans; 

    trans.set_command(cmd);
    trans.set_address(addr);
    trans.set_data_ptr(reinterpret_cast<unsigned char*>(data));
    trans.set_data_length(len);
    trans.set_streaming_width(len);
    trans.set_byte_enable_ptr(0);
    trans.set_dmi_allowed(false);
    trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    s->b_transport(trans, delay);
    if (trans.get_response_status() != tlm::TLM_OK_RESPONSE)
    {
      SC_REPORT_FATAL(name(), trans.get_response_string().c_str());
    }
    wait(delay);
  }

  void GetCount(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	unsigned int &c)
  {
    MemAccess(tlm::TLM_READ_COMMAND, s, addr, &c, sizeof(c));
  }

  void SetCount(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	unsigned int &c)
  {
    MemAccess(tlm::TLM_WRITE_COMMAND, s, addr, &c, sizeof(c));
  }

  void StoreFrame(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	long long &f)
  {
    MemAccess(tlm::TLM_WRITE_COMMAND, s, addr, &f, sizeof(f));
  }

  void PushVideoFrame(
	tlm::tlm_initiator_socket<> &s,
	long long &v)
  {
    GetCount(s, ADDRESS_VIDEO_RCVD_CNT, VframesRcvd);
    DEBUG_PRINTF("Video frames Sent-Rcvd = %u-%u = %u",
		VframesSent, VframesRcvd, VframesSent - VframesRcvd);
    while(VframesSent - VframesRcvd >= BUFFER_SIZE_VIDEO)
    { // wait for a free slot in the output buffer
      wait(POLLING_DELAY);
      GetCount(s, ADDRESS_VIDEO_RCVD_CNT, VframesRcvd);
      DEBUG_PRINTF("Video frames Sent-Rcvd = %u-%u = %u",
		VframesSent, VframesRcvd, VframesSent - VframesRcvd);
    }
    StoreFrame(s, ADDRESS_VIDEO_FRAME(VframesSent%BUFFER_SIZE_VIDEO), v);
    SetCount(s, ADDRESS_VIDEO_SENT_CNT, ++VframesSent);
  }

  void Video()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time VideoDelay(VIDEO_DELAY);
    sc_time t1(0, SC_US);

    VframesSent = 0;
    SetCount(MultiMediaStream, ADDRESS_VIDEO_SENT_CNT, VframesSent);
    wait(POLLING_DELAY);
    GetCount(MultiMediaStream, ADDRESS_VIDEO_RCVD_CNT, VframesRcvd);
    assert(VframesRcvd == 0);

    while(t1 < ClipLength)
    {
      t1 += VideoDelay;
      PushVideoFrame(MultiMediaStream, video);
    }
  }

  void PushAudioLeft(
	tlm::tlm_initiator_socket<> &s,
	long long &l)
  {
    GetCount(s, BASE_ADDRESS_AUDIO_L+OFFSET_AUDIO_RCVD_CNT, LframesRcvd);
    DEBUG_PRINTF("Left frames Sent-Rcvd = %u-%u = %u",
		LframesSent, LframesRcvd, LframesSent - LframesRcvd);
    while(LframesSent - LframesRcvd >= BUFFER_SIZE_AUDIO)
    { // wait for a free slot in the output buffer
      wait(POLLING_DELAY);
      GetCount(s, BASE_ADDRESS_AUDIO_L+OFFSET_AUDIO_RCVD_CNT, LframesRcvd);
      DEBUG_PRINTF("Left frames Sent-Rcvd = %u-%u = %u",
		LframesSent, LframesRcvd, LframesSent - LframesRcvd);
    }
    StoreFrame(s, BASE_ADDRESS_AUDIO_L+OFFSET_AUDIO_FRAME(
			LframesSent%BUFFER_SIZE_AUDIO), l);
    SetCount(s, BASE_ADDRESS_AUDIO_L+OFFSET_AUDIO_SENT_CNT, ++LframesSent);
  }

  void AudioL()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);

    LframesSent = 0;
    SetCount(MultiMediaStream, BASE_ADDRESS_AUDIO_L+OFFSET_AUDIO_SENT_CNT, LframesSent);
    wait(POLLING_DELAY);
    GetCount(MultiMediaStream, BASE_ADDRESS_AUDIO_L+OFFSET_AUDIO_RCVD_CNT, LframesRcvd);
    assert(LframesRcvd == 0);

    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      PushAudioLeft(MultiMediaStream, audioLeft);
    }
  }

  void PushAudioRight(
	tlm::tlm_initiator_socket<> &s,
	long long &r)
  {
    GetCount(s, BASE_ADDRESS_AUDIO_R+OFFSET_AUDIO_RCVD_CNT, RframesRcvd);
    DEBUG_PRINTF("Right frames Sent-Rcvd = %u-%u = %u",
		RframesSent, RframesRcvd, RframesSent - RframesRcvd);
    while(RframesSent - RframesRcvd >= BUFFER_SIZE_AUDIO)
    { // wait for a free slot in the output buffer
      wait(POLLING_DELAY);
      GetCount(s, BASE_ADDRESS_AUDIO_R+OFFSET_AUDIO_RCVD_CNT, RframesRcvd);
      DEBUG_PRINTF("Right frames Sent-Rcvd = %u-%u = %u",
		RframesSent, RframesRcvd, RframesSent - RframesRcvd);
    }
    StoreFrame(s, BASE_ADDRESS_AUDIO_R+OFFSET_AUDIO_FRAME(
			RframesSent%BUFFER_SIZE_AUDIO), r);
    SetCount(s, BASE_ADDRESS_AUDIO_R+OFFSET_AUDIO_SENT_CNT, ++RframesSent);
  }

  void AudioR()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);

    RframesSent = 0;
    SetCount(MultiMediaStream, BASE_ADDRESS_AUDIO_R+OFFSET_AUDIO_SENT_CNT, RframesSent);
    wait(POLLING_DELAY);
    GetCount(MultiMediaStream, BASE_ADDRESS_AUDIO_R+OFFSET_AUDIO_RCVD_CNT, RframesRcvd);
    assert(RframesRcvd == 0);

    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      PushAudioRight(MultiMediaStream, audioRight);
    }
  }

  SC_HAS_PROCESS(Stimulus);

  Stimulus(sc_module_name n)
  : sc_module(n)
  {
    video = 3;
    audioLeft = 27;
    audioRight = 15;
    SC_THREAD(Video);
    SC_THREAD(AudioL);
    SC_THREAD(AudioR);
  }
};

SC_MODULE(VideoCodec)
{
  tlm_utils::simple_initiator_socket<VideoCodec> s1;
  tlm_utils::simple_initiator_socket<VideoCodec> s2;

  unsigned int VframesSent, VframesRcvd;
  long long d, outFrame, inFrame;

  void MemAccess(
	tlm::tlm_command cmd,
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	void *data,
	unsigned int len)
  {
    sc_time delay = SC_ZERO_TIME;
    tlm::tlm_generic_payload trans; 

    trans.set_command(cmd);
    trans.set_address(addr);
    trans.set_data_ptr(reinterpret_cast<unsigned char*>(data));
    trans.set_data_length(len);
    trans.set_streaming_width(len);
    trans.set_byte_enable_ptr(0);
    trans.set_dmi_allowed(false);
    trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    s->b_transport(trans, delay);
    if (trans.get_response_status() != tlm::TLM_OK_RESPONSE)
    {
      SC_REPORT_FATAL(name(), trans.get_response_string().c_str());
    }
    wait(delay);
  }

  void GetCount(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	unsigned int &c)
  {
    MemAccess(tlm::TLM_READ_COMMAND, s, addr, &c, sizeof(c));
  }

  void SetCount(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	unsigned int &c)
  {
    MemAccess(tlm::TLM_WRITE_COMMAND, s, addr, &c, sizeof(c));
  }

  void LoadFrame(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	long long &f)
  {
    MemAccess(tlm::TLM_READ_COMMAND, s, addr, &f, sizeof(f));
  }

  void PopVideoFrame(
	tlm::tlm_initiator_socket<> &s,
	long long &v)
  {
    GetCount(s, ADDRESS_VIDEO_SENT_CNT, VframesSent);
    DEBUG_PRINTF("Frames Sent-Rcvd = %u-%u = %u",
		VframesSent, VframesRcvd, VframesSent - VframesRcvd);
    while(VframesSent - VframesRcvd == 0)
    { // wait for a filled slot in the input buffer
      wait(POLLING_DELAY);
      GetCount(s, ADDRESS_VIDEO_SENT_CNT, VframesSent);
      DEBUG_PRINTF("Frames Sent-Rcvd = %u-%u = %u",
		VframesSent, VframesRcvd, VframesSent - VframesRcvd);
    }
    LoadFrame(s, ADDRESS_VIDEO_FRAME(VframesRcvd%BUFFER_SIZE_VIDEO), v);
    SetCount(s, ADDRESS_VIDEO_RCVD_CNT, ++VframesRcvd);
  }

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void DevWrite(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	void *data,
	unsigned int len)
  {
    sc_time delay = SC_ZERO_TIME;
    tlm::tlm_generic_payload trans; 

    trans.set_command(tlm::TLM_WRITE_COMMAND);
    trans.set_address(addr);
    trans.set_data_ptr(reinterpret_cast<unsigned char*>(data));
    trans.set_data_length(len);
    trans.set_streaming_width(len);
    trans.set_byte_enable_ptr(0);
    trans.set_dmi_allowed(false);
    trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    s->b_transport(trans, delay);
    if (trans.get_response_status() != tlm::TLM_OK_RESPONSE)
    {
      SC_REPORT_FATAL(name(), trans.get_response_string().c_str());
    }
    wait(delay);
  }

  void main(void)
  {
    d = 1;

    VframesRcvd = 0;
    SetCount(s1, ADDRESS_VIDEO_RCVD_CNT, VframesRcvd);
    wait(POLLING_DELAY);
    GetCount(s1, ADDRESS_VIDEO_SENT_CNT, VframesSent);

    while(1)
    {
      PopVideoFrame(s1, inFrame);
      wait(HALF_VIDEO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_VIDEO_DELAY);
      DevWrite(s2, 0, &outFrame, sizeof(outFrame));
    }
  }

  SC_HAS_PROCESS(VideoCodec);

  VideoCodec(sc_module_name n)
  : sc_module(n)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(AudioCodec)
{
  tlm_utils::simple_initiator_socket<AudioCodec> s1;
  tlm_utils::simple_initiator_socket<AudioCodec> s2;

  const sc_dt::uint64 BaseAddress;
  unsigned int AframesSent, AframesRcvd;
  long long d, inFrame, outFrame;

  void MemAccess(
	tlm::tlm_command cmd,
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	void *data,
	unsigned int len)
  {
    sc_time delay = SC_ZERO_TIME;
    tlm::tlm_generic_payload trans; 

    trans.set_command(cmd);
    trans.set_address(addr);
    trans.set_data_ptr(reinterpret_cast<unsigned char*>(data));
    trans.set_data_length(len);
    trans.set_streaming_width(len);
    trans.set_byte_enable_ptr(0);
    trans.set_dmi_allowed(false);
    trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    s->b_transport(trans, delay);
    if (trans.get_response_status() != tlm::TLM_OK_RESPONSE)
    {
      SC_REPORT_FATAL(name(), trans.get_response_string().c_str());
    }
    wait(delay);
  }

  void GetCount(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	unsigned int &c)
  {
    MemAccess(tlm::TLM_READ_COMMAND, s, addr, &c, sizeof(c));
  }

  void SetCount(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	unsigned int &c)
  {
    MemAccess(tlm::TLM_WRITE_COMMAND, s, addr, &c, sizeof(c));
  }

  void LoadFrame(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	long long &f)
  {
    MemAccess(tlm::TLM_READ_COMMAND, s, addr, &f, sizeof(f));
  }

  void PopAudioFrame(
	tlm::tlm_initiator_socket<> &s,
	long long &a)
  {
    GetCount(s, BaseAddress+OFFSET_AUDIO_SENT_CNT, AframesSent);
    DEBUG_PRINTF("Frames Sent-Rcvd = %u-%u = %u",
		AframesSent, AframesRcvd, AframesSent - AframesRcvd);
    while(AframesSent - AframesRcvd == 0)
    { // wait for a filled slot in the input buffer
      wait(POLLING_DELAY);
      GetCount(s, BaseAddress+OFFSET_AUDIO_SENT_CNT, AframesSent);
      DEBUG_PRINTF("Frames Sent-Rcvd = %u-%u = %u",
		AframesSent, AframesRcvd, AframesSent - AframesRcvd);
    }
    LoadFrame(s, BaseAddress+OFFSET_AUDIO_FRAME(AframesRcvd%BUFFER_SIZE_AUDIO), a);
    SetCount(s, BaseAddress+OFFSET_AUDIO_RCVD_CNT, ++AframesRcvd);
  }

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void DevWrite(
	tlm::tlm_initiator_socket<> &s,
	sc_dt::uint64 addr,
	void *data,
	unsigned int len)
  {
    sc_time delay = SC_ZERO_TIME;
    tlm::tlm_generic_payload trans; 

    trans.set_command(tlm::TLM_WRITE_COMMAND);
    trans.set_address(addr);
    trans.set_data_ptr(reinterpret_cast<unsigned char*>(data));
    trans.set_data_length(len);
    trans.set_streaming_width(len);
    trans.set_byte_enable_ptr(0);
    trans.set_dmi_allowed(false);
    trans.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    s->b_transport(trans, delay);
    if (trans.get_response_status() != tlm::TLM_OK_RESPONSE)
    {
      SC_REPORT_FATAL(name(), trans.get_response_string().c_str());
    }
    wait(delay);
  }

  void main(void)
  {
    d = 1;

    AframesRcvd = 0;
    SetCount(s1, BaseAddress+OFFSET_AUDIO_RCVD_CNT, AframesRcvd);
    wait(POLLING_DELAY);
    GetCount(s1, BaseAddress+OFFSET_AUDIO_SENT_CNT, AframesSent);

    while(1)
    {
      PopAudioFrame(s1, inFrame);
      wait(HALF_AUDIO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_AUDIO_DELAY);
      DevWrite(s2, 0, &outFrame, sizeof(outFrame));
    }
  }

  SC_HAS_PROCESS(AudioCodec);

  AudioCodec(sc_module_name n,
           const sc_dt::uint64 BaseAddress)
  : sc_module(n)
  , BaseAddress(BaseAddress)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(DUT)
{
  tlm::tlm_target_socket< > MultiMediaStream;
  tlm::tlm_initiator_socket< > VideoOut;
  tlm::tlm_initiator_socket< > AudioOutL;
  tlm::tlm_initiator_socket< > AudioOutR;

  Memory4p   memory;
  VideoCodec video;
  AudioCodec left, right;

  DUT(sc_module_name n)
  : sc_module(n)
  , memory("memory")
  , video("video")
  , left("left", BASE_ADDRESS_AUDIO_L)
  , right("right", BASE_ADDRESS_AUDIO_R)
  { // note: hierarchical socket binding must happen in constructor
    // for target sockets, parent should be bound to child
    MultiMediaStream.bind(memory.s0);
    video.s1.bind(memory.s1);
    left.s1.bind(memory.s2);
    right.s1.bind(memory.s3);
    // for initiator sockets, child should be bound to parent
    video.s2.bind(VideoOut);
    left.s2.bind(AudioOutL);
    right.s2.bind(AudioOutR);
  }
};

SC_MODULE(Display)
{
  tlm_utils::simple_target_socket<Display> VideoIn;

  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesDisplayed;
#endif
  sc_time TimePlayed;

  void display(long long
#if (defined CHECK_FRAMES || defined DUMP_FRAMES)
               Frame
#endif
               )
  {
#ifdef PRINT_FRAMES
    putchar('V');       // display the frame
    fflush(stdout);
#endif
#ifdef DUMP_FRAMES
    fprintf(fh[0], "  %20lldLL,\n", Frame);
#endif
#ifdef COUNT_FRAMES
#  ifdef CHECK_FRAMES
    if (FramesDisplayed < sizeof(vframes)/sizeof(long long))
    { assert(Frame == vframes[FramesDisplayed]); }
#  endif
    FramesDisplayed++;
#endif
  }

  void quit()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(VIDEO_DELAY);
    TimePlayed += AudioDelay;
    if (TimePlayed >= ClipLength)
    {
      wait(); // starve this thread so that simulation can quit
    }
  }

  void receive(tlm::tlm_generic_payload& trans, sc_time& delay)
  {
    tlm::tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    unsigned char   *ptr = trans.get_data_ptr();
    unsigned int     len = trans.get_data_length();
    unsigned int     wid = trans.get_streaming_width();
    unsigned char   *byt = trans.get_byte_enable_ptr();
    if (  (cmd == tlm::TLM_WRITE_COMMAND)
        &&(adr == 0)
        &&(ptr != 0)
        &&(len == sizeof(Frame))
        &&(wid == sizeof(Frame))
        &&(byt == 0))
    {
      memcpy(&Frame, ptr, len);
      trans.set_response_status(tlm::TLM_OK_RESPONSE);
      delay += sc_time(DEV_WRITE_ACCESS_DELAY);
    }
    else
    {
      trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
      return;
    }
    display(Frame);
    quit();
  }

  SC_CTOR(Display)
  {
#ifdef COUNT_FRAMES
    FramesDisplayed = 0;
#endif
    TimePlayed = SC_ZERO_TIME;
    VideoIn.register_b_transport(this, &Display::receive);
  }
};

SC_MODULE(Speaker)
{
  tlm_utils::simple_target_socket<Speaker> AudioIn;

  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesPlayed;
#endif
  sc_time TimePlayed;
  char Channel;

  void sound(long long
#if (defined CHECK_FRAMES || defined DUMP_FRAMES)
               Frame
#endif
               )
  {
#ifdef PRINT_FRAMES
    putchar(Channel); // output the sound
    fflush(stdout);
#endif
#ifdef DUMP_FRAMES
    fprintf(fh[Channel=='L'?1:2], "  %20lldLL,\n", Frame);
#endif
#ifdef COUNT_FRAMES
#  ifdef CHECK_FRAMES
    if (Channel == 'L')
    { if (FramesPlayed < sizeof(lframes)/sizeof(long long))
      { assert(Frame == lframes[FramesPlayed]); }
    }
    else
    { if (FramesPlayed < sizeof(rframes)/sizeof(long long))
      { assert(Frame == rframes[FramesPlayed]); }
    }
#  endif
    FramesPlayed++;
#endif
  }

  void quit()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    TimePlayed += AudioDelay;
    if (TimePlayed >= ClipLength)
    {
      wait(); // starve this thread so that simulation can quit
    }
  }

  void receive(tlm::tlm_generic_payload& trans, sc_time& delay)
  {
    tlm::tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    unsigned char   *ptr = trans.get_data_ptr();
    unsigned int     len = trans.get_data_length();
    unsigned int     wid = trans.get_streaming_width();
    unsigned char   *byt = trans.get_byte_enable_ptr();
    if (  (cmd == tlm::TLM_WRITE_COMMAND)
        &&(adr == 0)
        &&(ptr != 0)
        &&(len == sizeof(Frame))
        &&(wid == sizeof(Frame))
        &&(byt == 0))
    {
      memcpy(&Frame, ptr, len);
      trans.set_response_status(tlm::TLM_OK_RESPONSE);
      delay += sc_time(DEV_WRITE_ACCESS_DELAY);
    }
    else
    {
      trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
      return;
    }
    sound(Frame);
    quit();
  }

  Speaker(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
#ifdef COUNT_FRAMES
    FramesPlayed = 0;
#endif
    TimePlayed = SC_ZERO_TIME;
    AudioIn.register_b_transport(this, &Speaker::receive);
  }
};

SC_MODULE(Monitor)
{
  tlm::tlm_target_socket< > VideoIn;
  tlm::tlm_target_socket< > AudioInL;
  tlm::tlm_target_socket< > AudioInR;

  Display screen;
  Speaker speakerL, speakerR;

  SC_CTOR(Monitor):
        screen("screen"),
        speakerL("speakerL", 'L'),
        speakerR("speakerR", 'R')
  { // note: hierarchical socket binding must happen in constructor
    // for target sockets, parent should be bound to child
    // (see page 461 of the IEEE Std. 1666-2011, item m)
    VideoIn.bind(screen.VideoIn);
    AudioInL.bind(speakerL.AudioIn);
    AudioInR.bind(speakerR.AudioIn);
  }
};

SC_MODULE(Top)
{
  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;

  SC_CTOR(Top)
  : stimulus("stimulus")
  , dut("dut")
  , monitor("monitor")
  { // note: hierarchical socket binding must happen in constructor
    stimulus.MultiMediaStream.bind(dut.MultiMediaStream);
    dut.VideoOut.bind(monitor.VideoIn);
    dut.AudioOutL.bind(monitor.AudioInL);
    dut.AudioOutR.bind(monitor.AudioInR);
  }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");
  sc_time	ClipLength(CLIP_LENGTH);
  sc_time	VideoDelay(VIDEO_DELAY);
  sc_time	AudioDelay(AUDIO_DELAY);
#ifdef TRACE_FILE
  sc_trace_file *f = sc_create_vcd_trace_file(argv[0]);
  f->set_time_unit(sc_time(10, SC_US).to_seconds(), SC_SEC);
  sc_trace(f, top.dut.video.d, "top.dut.video.d");
  sc_trace(f, top.dut.left.d, "top.dut.left.d");
  sc_trace(f, top.dut.right.d, "top.dut.right.d");
#  ifdef COUNT_FRAMES
  sc_trace(f, top.monitor.screen.FramesDisplayed, "top.monitor.screen.FramesDisplayed");
  sc_trace(f, top.monitor.speakerL.FramesPlayed, "top.monitor.speakerL.FramesPlayed");
  sc_trace(f, top.monitor.speakerR.FramesPlayed, "top.monitor.speakerR.FramesPlayed");
#  endif
#endif

#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
#endif
  assert(ADDRESS_VIDEO_FRAME(BUFFER_SIZE_VIDEO) <= BASE_ADDRESS_AUDIO_L);
  assert(BASE_ADDRESS_AUDIO_L + OFFSET_AUDIO_FRAME(BUFFER_SIZE_AUDIO)
	<= BASE_ADDRESS_AUDIO_R);
  assert(BASE_ADDRESS_AUDIO_R + OFFSET_AUDIO_FRAME(BUFFER_SIZE_AUDIO)
	<= MEMORY_SIZE);
  printf("Playing...\n");
#ifdef DUMP_FRAMES
  printf("Dumping v.frames, l.frames, r.frames...\n");
  if (!(fh[0] = fopen("v.frames","w"))) return 1;
  if (!(fh[1] = fopen("l.frames","w"))) return 2;
  if (!(fh[2] = fopen("r.frames","w"))) return 3;
  fprintf(fh[0], "long long vframes[] = {\n");
  fprintf(fh[1], "long long lframes[] = {\n");
  fprintf(fh[2], "long long rframes[] = {\n");
#endif
#ifdef CHECK_FRAMES
  printf("Checking v.frames, l.frames, r.frames...\n");
#endif
  sc_start();
#ifdef PRINT_FRAMES
  printf("\n");
#endif
#ifdef COUNT_FRAMES
  printf("%u video frames displayed\n", top.monitor.screen.FramesDisplayed);
  printf("%u left audio frames played\n", top.monitor.speakerL.FramesPlayed);
  printf("%u right audio frames played\n", top.monitor.speakerR.FramesPlayed);
  assert(top.monitor.screen.FramesDisplayed ==
		ClipLength.value()/VideoDelay.value()+1);
  assert(top.monitor.speakerL.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
  assert(top.monitor.speakerR.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
#endif
#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
  assert(sc_time_stamp() >= ClipLength);
#endif
#ifdef TRACE_FILE
  sc_close_vcd_trace_file(f);
#endif
#ifdef DUMP_FRAMES
  fprintf(fh[0], "  };\n");
  fprintf(fh[1], "  };\n");
  fprintf(fh[2], "  };\n");
  fclose(fh[0]);
  fclose(fh[1]);
  fclose(fh[2]);
#endif
  printf("Done!\n");
  return 0;
}

// EOF
