// play_TLM2_memX.cpp: conceptual audio/video stream decoding example
// Author: RISC project members, CECS
//
// 02/04/19 RD  version with 1 memory, cross-hierarchy socket binding
// 02/03/19 RD  added validation of display and speaker data
// 01/29/19 RD  version with 3 memories, cross-hierarchy socket binding
// 01/28/18 RD  improved generic payload to meet Doulos recommendations
//              and applied some cleaning for better presentation
// 27/11/18 ZC  use sockets to replace ports and channels
// 06/05/18 RD  confirmed clean sc_time constants with GCC 6.3.1
// 04/27/18 RD  clean sc_time constants for newer compilers (e.g. GCC 4.9.2)
// 04/26/18 RD  merged tracing version into main file (define TRACE_FILE)
// 04/24/18 DM  added tracing of stream data (sc_trace is now supported)
// 07/10/17 RD  split the delay into two parts (as suggested by ZC)
// 04/05/17 RD  removed sc_stop from stimulus, let monitor terminate;
//              added COUNT_FRAMES to ensure proper completion
// 03/08/17 RD  avoid duplicated DELAY macros, set BUFSIZE to 10
// 01/13/17 GL  separate version play_fifo.cpp with sc_fifo channels
// 10/04/16 RD  removed obsolete C artifacts from the user-defined queue channel
// 09/30/16 RD  cleanup for release 0.3.0
// 03/07/15 RD  bug fix: use sc_stop() to quit cleanly (exit() is not MT-safe)
// 10/30/15 RD  polished for release 0.2.1
// 09/04/15 GL  use 'long long' (not 'double') to avoid floating-point overflow
//              (FPU exception handling shows high performance variation;
//              moreover, integer arithmetic is more realistic for decoder)
// 08/30/15 GL  initial version based on SpecC example play.sc by RD

#include <stdio.h>
#include "systemc.h"
#include "tlm.h"                               // TLM headers
#include "tlm_utils/simple_initiator_socket.h"
#include "tlm_utils/simple_target_socket.h"
#include <string>
using namespace sc_core;
using namespace std;

// frame rates
#define AUDIO_DELAY     26120, SC_US
#define VIDEO_DELAY     33330, SC_US
#define HALF_AUDIO_DELAY 26120/2, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US

// simulated time
#define CLIP_LENGTH     10, SC_SEC

// simulated workload unit
#define AUDIOLOAD        7500000
#define VIDEOLOAD       10000000

// illustrate and check frame decoding
//#define PRINT_TIME
//#define PRINT_FRAMES
//#define COUNT_FRAMES
//#define TRACE_FILE
//#define DUMP_FRAMES
//#define CHECK_FRAMES

#ifdef DUMP_FRAMES
FILE *fh[3];
#endif
#ifdef CHECK_FRAMES
#  include "v.frames"
#  include "l.frames"
#  include "r.frames"
#endif

// memory sizes and segmentation
#define MEMORY_SIZE     0x10000
#define ADDRESS_VIDEO   0x00000
#define ADDRESS_AUDIO_L 0x01000
#define ADDRESS_AUDIO_R 0x01800

SC_MODULE(Memory4p)
{
  public:
    tlm_utils::simple_target_socket<Memory4p>  s0;
    tlm_utils::simple_target_socket<Memory4p>  s1;
    tlm_utils::simple_target_socket<Memory4p>  s2;
    tlm_utils::simple_target_socket<Memory4p>  s3;

    SC_CTOR(Memory4p)
    {
      M = new unsigned char[MEMORY_SIZE];
      s0.register_b_transport(this, &Memory4p::access);
      s1.register_b_transport(this, &Memory4p::access);
      s2.register_b_transport(this, &Memory4p::access);
      s3.register_b_transport(this, &Memory4p::access);
    }

    void access(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      tlm::tlm_command cmd = trans.get_command();
      sc_dt::uint64    adr = trans.get_address();
      unsigned char   *ptr = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      unsigned int     wid = trans.get_streaming_width();
      unsigned char   *byt = trans.get_byte_enable_ptr();
      if (  (adr + len <= MEMORY_SIZE)
          &&(ptr != 0)
          &&(wid == len)
          &&(byt == 0))
      {
        if (cmd == tlm::TLM_READ_COMMAND)
        {
          memcpy(ptr, &M[adr], len);
          trans.set_response_status(tlm::TLM_OK_RESPONSE);
        }
        else if (cmd == tlm::TLM_WRITE_COMMAND)
        {
          memcpy(&M[adr], ptr, len);
          trans.set_response_status(tlm::TLM_OK_RESPONSE);
        }
        else
        {
          trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
          return;
        }
      }
      else
      {
        trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
        return;
      }
    }

  private:
    unsigned char *M;
};

SC_MODULE(Stimulus)
{
  tlm_utils::simple_initiator_socket<Stimulus> MultiMediaStream;

  sc_event &VCodecStart, &VCodecDone;
  sc_event &LCodecStart, &LCodecDone;
  sc_event &RCodecStart, &RCodecDone;

  long long video;
  long long audioLeft;
  long long audioRight;

  void Video()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time VideoDelay(VIDEO_DELAY);
    sc_time t1(0, SC_US);

    while(t1 < ClipLength)
    {
      t1 += VideoDelay;
      sc_time delay0 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans0; 
      trans0.set_command(tlm::TLM_WRITE_COMMAND);
      trans0.set_address(ADDRESS_VIDEO);
      trans0.set_data_ptr(reinterpret_cast<unsigned char*>(&video));
      trans0.set_data_length(sizeof(video));
      trans0.set_streaming_width(sizeof(video));
      trans0.set_byte_enable_ptr(0);
      trans0.set_dmi_allowed(false);
      trans0.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      MultiMediaStream->b_transport(trans0, delay0);
      if (trans0.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans0.get_response_string().c_str());
      }
      VCodecStart.notify(SC_ZERO_TIME);
      wait(VCodecDone);
    }
  }

  void AudioL()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);

    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      sc_time delay1 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans1;
      trans1.set_command(tlm::TLM_WRITE_COMMAND);
      trans1.set_address(ADDRESS_AUDIO_L);
      trans1.set_data_ptr(reinterpret_cast<unsigned char*>(&audioLeft));
      trans1.set_data_length(sizeof(audioLeft));
      trans1.set_streaming_width(sizeof(audioLeft));
      trans1.set_byte_enable_ptr(0);
      trans1.set_dmi_allowed(false);
      trans1.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      MultiMediaStream->b_transport(trans1, delay1);
      if (trans1.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans1.get_response_string().c_str());
      }
      LCodecStart.notify(SC_ZERO_TIME);
      wait(LCodecDone);
    }
  }

  void AudioR()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);

    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      sc_time delay2 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans2; 
      trans2.set_command(tlm::TLM_WRITE_COMMAND);
      trans2.set_address(ADDRESS_AUDIO_R);
      trans2.set_data_ptr(reinterpret_cast<unsigned char*>(&audioRight));
      trans2.set_data_length(sizeof(audioRight));
      trans2.set_streaming_width(sizeof(audioRight));
      trans2.set_byte_enable_ptr(0);
      trans2.set_dmi_allowed(false);
      trans2.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      MultiMediaStream->b_transport(trans2, delay2);
      if (trans2.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans2.get_response_string().c_str());
      }
      RCodecStart.notify(SC_ZERO_TIME);
      wait(RCodecDone);
    }
  }

  SC_HAS_PROCESS(Stimulus);

  Stimulus(sc_module_name n,
           sc_event &Vstart, sc_event &Vdone,
           sc_event &Lstart, sc_event &Ldone,
           sc_event &Rstart, sc_event &Rdone)
  : sc_module(n)
  , VCodecStart(Vstart), VCodecDone(Vdone)
  , LCodecStart(Lstart), LCodecDone(Ldone)
  , RCodecStart(Rstart), RCodecDone(Rdone)
  {
    video = 3;
    audioLeft = 27;
    audioRight = 15;
    SC_THREAD(Video);
    SC_THREAD(AudioL);
    SC_THREAD(AudioR);
  }
};

SC_MODULE(VideoCodec)
{
  tlm_utils::simple_initiator_socket<VideoCodec> s1;
  tlm_utils::simple_initiator_socket<VideoCodec> s2;

  sc_event &VCodecStart, &VCodecDone;

  long long d, outFrame, inFrame;
  
  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    d = 1;

    while(1)
    {
      wait(VCodecStart);

      sc_time delay0 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans0; 
      trans0.set_command(tlm::TLM_READ_COMMAND);
      trans0.set_address(ADDRESS_VIDEO);
      trans0.set_data_ptr(reinterpret_cast<unsigned char*>(&inFrame));
      trans0.set_data_length(sizeof(inFrame));
      trans0.set_streaming_width(sizeof(inFrame));
      trans0.set_byte_enable_ptr(0);
      trans0.set_dmi_allowed(false);
      trans0.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      s1->b_transport(trans0, delay0);
      if (trans0.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans0.get_response_string().c_str());
      }

      wait(HALF_VIDEO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_VIDEO_DELAY);

      sc_time delay1 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans1; 
      trans1.set_command(tlm::TLM_WRITE_COMMAND);
      trans1.set_address(0);
      trans1.set_data_ptr(reinterpret_cast<unsigned char*>(&outFrame));
      trans1.set_data_length(sizeof(outFrame));
      trans1.set_streaming_width(sizeof(outFrame));
      trans1.set_byte_enable_ptr(0);
      trans1.set_dmi_allowed(false);
      trans1.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      s2->b_transport(trans1, delay1);
      if (trans1.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans1.get_response_string().c_str());
      }

      VCodecDone.notify(SC_ZERO_TIME);
    }
  }

  SC_HAS_PROCESS(VideoCodec);

  VideoCodec(sc_module_name n,
           sc_event &Vstart, sc_event &Vdone)
  : sc_module(n)
  , VCodecStart(Vstart), VCodecDone(Vdone)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(AudioCodec)
{
  tlm_utils::simple_initiator_socket<AudioCodec> s1;
  tlm_utils::simple_initiator_socket<AudioCodec> s2;

  sc_event &ACodecStart, &ACodecDone;
  const sc_dt::uint64 DataInAddr;

  long long d, inFrame, outFrame;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    d = 1;

    while(1)
    {
      wait(ACodecStart);

      sc_time delay0 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans0; 
      trans0.set_command(tlm::TLM_READ_COMMAND);
      trans0.set_address(DataInAddr);
      trans0.set_data_ptr(reinterpret_cast<unsigned char*>(&inFrame));
      trans0.set_data_length(sizeof(inFrame));
      trans0.set_streaming_width(sizeof(inFrame));
      trans0.set_byte_enable_ptr(0);
      trans0.set_dmi_allowed(false);
      trans0.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      s1->b_transport(trans0, delay0);
      if (trans0.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans0.get_response_string().c_str());
      }

      wait(HALF_AUDIO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_AUDIO_DELAY);

      sc_time delay1 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans1; 
      trans1.set_command(tlm::TLM_WRITE_COMMAND);
      trans1.set_address(0);
      trans1.set_data_ptr(reinterpret_cast<unsigned char*>(&outFrame));
      trans1.set_data_length(sizeof(outFrame));
      trans1.set_streaming_width(sizeof(outFrame));
      trans1.set_byte_enable_ptr(0);
      trans1.set_dmi_allowed(false);
      trans1.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      s2->b_transport(trans1, delay1);
      if (trans1.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans1.get_response_string().c_str());
      }

      ACodecDone.notify(SC_ZERO_TIME);
    }
  }

  SC_HAS_PROCESS(AudioCodec);

  AudioCodec(sc_module_name n,
           sc_event &Astart, sc_event &Adone,
           const sc_dt::uint64 DataInAddr)
  : sc_module(n)
  , ACodecStart(Astart), ACodecDone(Adone)
  , DataInAddr(DataInAddr)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(DUT)
{
  Memory4p   memory;
  VideoCodec video;
  AudioCodec left, right;

  DUT(sc_module_name n,
           sc_event &Vstart, sc_event &Vdone,
           sc_event &Lstart, sc_event &Ldone,
           sc_event &Rstart, sc_event &Rdone)
  : sc_module(n)
  , memory("memory")
  , video("video", Vstart, Vdone)
  , left("left", Lstart, Ldone, ADDRESS_AUDIO_L)
  , right("right", Rstart, Rdone, ADDRESS_AUDIO_R)
  { }

  void before_end_of_elaboration()
  {
    video.s1.bind(memory.s1);
    left.s1.bind(memory.s2);
    right.s1.bind(memory.s3);
  }
};

SC_MODULE(Display)
{
  tlm_utils::simple_target_socket<Display> VideoIn;

  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesDisplayed;
#endif

  void display(long long
#if (defined CHECK_FRAMES || defined DUMP_FRAMES)
               Frame
#endif
               )
  {
#ifdef PRINT_FRAMES
    putchar('V');       // display the frame
    fflush(stdout);
#endif
#ifdef DUMP_FRAMES
    fprintf(fh[0], "  %20lldLL,\n", Frame);
#endif
#ifdef COUNT_FRAMES
#  ifdef CHECK_FRAMES
    if (FramesDisplayed < sizeof(vframes)/sizeof(long long))
    { assert(Frame == vframes[FramesDisplayed]); }
#  endif
    FramesDisplayed++;
#endif
  }

  void receive(tlm::tlm_generic_payload& trans, sc_time& delay)
  {
    tlm::tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    unsigned char   *ptr = trans.get_data_ptr();
    unsigned int     len = trans.get_data_length();
    unsigned int     wid = trans.get_streaming_width();
    unsigned char   *byt = trans.get_byte_enable_ptr();
    if (  (cmd == tlm::TLM_WRITE_COMMAND)
        &&(adr == 0)
        &&(ptr != 0)
        &&(len == sizeof(Frame))
        &&(wid == sizeof(Frame))
        &&(byt == 0))
    {
      memcpy(&Frame, ptr, len);
      trans.set_response_status(tlm::TLM_OK_RESPONSE);
    }
    else
    {
      trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
      return;
    }
    display(Frame);
  }

  SC_CTOR(Display)
  {
#ifdef COUNT_FRAMES
    FramesDisplayed = 0;
#endif
    VideoIn.register_b_transport(this, &Display::receive);
  }
};

SC_MODULE(Speaker)
{
  tlm_utils::simple_target_socket<Speaker> AudioIn;

  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesPlayed;
#endif
  char Channel;

  void sound(long long
#if (defined CHECK_FRAMES || defined DUMP_FRAMES)
               Frame
#endif
               )
  {
#ifdef PRINT_FRAMES
    putchar(Channel); // output the sound
    fflush(stdout);
#endif
#ifdef DUMP_FRAMES
    fprintf(fh[Channel=='L'?1:2], "  %20lldLL,\n", Frame);
#endif
#ifdef COUNT_FRAMES
#  ifdef CHECK_FRAMES
    if (Channel == 'L')
    { if (FramesPlayed < sizeof(lframes)/sizeof(long long))
      { assert(Frame == lframes[FramesPlayed]); }
    }
    else
    { if (FramesPlayed < sizeof(rframes)/sizeof(long long))
      { assert(Frame == rframes[FramesPlayed]); }
    }
#  endif
    FramesPlayed++;
#endif
  }

  void receive(tlm::tlm_generic_payload& trans, sc_time& delay)
  {
    tlm::tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    unsigned char   *ptr = trans.get_data_ptr();
    unsigned int     len = trans.get_data_length();
    unsigned int     wid = trans.get_streaming_width();
    unsigned char   *byt = trans.get_byte_enable_ptr();
    if (  (cmd == tlm::TLM_WRITE_COMMAND)
        &&(adr == 0)
        &&(ptr != 0)
        &&(len == sizeof(Frame))
        &&(wid == sizeof(Frame))
        &&(byt == 0))
    {
      memcpy(&Frame, ptr, len);
      trans.set_response_status(tlm::TLM_OK_RESPONSE);
    }
    else
    {
      trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
      return;
    }
    sound(Frame);
  }

  Speaker(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
#ifdef COUNT_FRAMES
    FramesPlayed = 0;
#endif
    AudioIn.register_b_transport(this, &Speaker::receive);
  }
};

SC_MODULE(Monitor)
{
  Display screen;
  Speaker speakerL, speakerR;

  SC_CTOR(Monitor):
        screen("screen"),
        speakerL("speakerL", 'L'),
        speakerR("speakerR", 'R')
  { }
};

SC_MODULE(Top)
{
  sc_event VCodecStart, VCodecDone;
  sc_event LCodecStart, LCodecDone;
  sc_event RCodecStart, RCodecDone;

  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;

  SC_CTOR(Top)
  : stimulus("stimulus",
        VCodecStart, VCodecDone,
        LCodecStart, LCodecDone,
        RCodecStart, RCodecDone)
  , dut("dut",
        VCodecStart, VCodecDone,
        LCodecStart, LCodecDone,
        RCodecStart, RCodecDone)
  , monitor("monitor")
  { }

  void before_end_of_elaboration()
  {
    stimulus.MultiMediaStream.bind(dut.memory.s0);

    dut.video.s2.bind(monitor.screen.VideoIn);
    dut.left.s2.bind(monitor.speakerL.AudioIn);
    dut.right.s2.bind(monitor.speakerR.AudioIn);
  }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");
  sc_time	ClipLength(CLIP_LENGTH);
  sc_time	VideoDelay(VIDEO_DELAY);
  sc_time	AudioDelay(AUDIO_DELAY);
#ifdef TRACE_FILE
  sc_trace_file *f = sc_create_vcd_trace_file(argv[0]);
  f->set_time_unit(sc_time(10, SC_US).to_seconds(), SC_SEC);
  sc_trace(f, top.dut.video.d, "top.dut.video.d");
  sc_trace(f, top.dut.left.d, "top.dut.left.d");
  sc_trace(f, top.dut.right.d, "top.dut.right.d");
#  ifdef COUNT_FRAMES
  sc_trace(f, top.monitor.screen.FramesDisplayed, "top.monitor.screen.FramesDisplayed");
  sc_trace(f, top.monitor.speakerL.FramesPlayed, "top.monitor.speakerL.FramesPlayed");
  sc_trace(f, top.monitor.speakerR.FramesPlayed, "top.monitor.speakerR.FramesPlayed");
#  endif
#endif

#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
#endif
  printf("Playing...\n");
#ifdef DUMP_FRAMES
  printf("Dumping v.frames, l.frames, r.frames...\n");
  if (!(fh[0] = fopen("v.frames","w"))) return 1;
  if (!(fh[1] = fopen("l.frames","w"))) return 2;
  if (!(fh[2] = fopen("r.frames","w"))) return 3;
  fprintf(fh[0], "long long vframes[] = {\n");
  fprintf(fh[1], "long long lframes[] = {\n");
  fprintf(fh[2], "long long rframes[] = {\n");
#endif
#ifdef CHECK_FRAMES
  printf("Checking v.frames, l.frames, r.frames...\n");
#endif
  sc_start();
#ifdef PRINT_FRAMES
  printf("\n");
#endif
#ifdef COUNT_FRAMES
  printf("%u video frames displayed\n", top.monitor.screen.FramesDisplayed);
  printf("%u left audio frames played\n", top.monitor.speakerL.FramesPlayed);
  printf("%u right audio frames played\n", top.monitor.speakerR.FramesPlayed);
  assert(top.monitor.screen.FramesDisplayed ==
		ClipLength.value()/VideoDelay.value()+1);
  assert(top.monitor.speakerL.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
  assert(top.monitor.speakerR.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
#endif
#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
  assert(sc_time_stamp() >= ClipLength);
#endif
#ifdef TRACE_FILE
  sc_close_vcd_trace_file(f);
#endif
#ifdef DUMP_FRAMES
  fprintf(fh[0], "  };\n");
  fprintf(fh[1], "  };\n");
  fprintf(fh[2], "  };\n");
  fclose(fh[0]);
  fclose(fh[1]);
  fclose(fh[2]);
#endif
  printf("Done!\n");
  return 0;
}

// EOF
