#include "port_mapping.h"

#include "design.h"
#include "hierarchical_channel_instance.h"
#include "instance.h"
#include "module.h"
#include "port.h"
#include "socket.h"
#include "module_instance.h"
#include "primitive_channel_instance.h"

#include "../tools/class_is_derived_from.h"
#include "../tools/misc.h"
#include "../tools/function_helper.h"

std::list<risc::InstanceTree>
risc::get_mapped_channel_instances(Port *port)
{
  // This list stores all channels which are mapped to the given port.
  // The port can occur in many instance trees, so, one port can be bound
  // to many channels
  std::list<InstanceTree> mapped_channels;

  // here we are searching for all instance trees which include this port
  InstanceTree root = risc::get_design()->get_root();
  std::list<InstanceTree> port_instances
    = root.find_all_instances_of(risc::tools::get_sym_from_def(port->get_ast_node()));

  // Iteration over all port instance trees
  for(std::list<InstanceTree>::iterator
      iter  = port_instances.begin();
      iter != port_instances.end();
      iter++) {

    mapped_channels.push_back(get_mapped_channel_instance(*iter));
  }

  return mapped_channels;
}


SgVariableSymbol*
risc::get_mapped_symbol(
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable)
{
  /* arguments:
  in parent_module, there is
    parent_variable.port.bind(...)
  or
    parent_variable.port(...)
  */ 

  //this not only works for ports, but also for sockets
  //the port_mapping_style1,2,3 work for both ports 
  //and sockets bindings
  
  //---following added for PSG support
  //if the port mapping for parent_variable.port is already mapped
  //return mapped symbol
  risc::Design* d = risc::get_design();
  for(ModuleInstanceVector::iterator mi_iter = d->loaded_module_instances.begin();
    mi_iter != d->loaded_module_instances.end();
    mi_iter ++)
  {
    if((*mi_iter)->get_ast_node() == parent_variable) 
    {
      // module instance already loaded
      // then we need to find the port
      // and its mapping
      for(std::map<Port*, Object*>::iterator mapping_iter = (*mi_iter)->mapping_.begin();
        mapping_iter!=(*mi_iter)->mapping_.end();
        mapping_iter++)
      {
        
        if(((*mapping_iter).first)->get_symbol()==port) 
          // found the port
          // return the mapping
          return ((*mapping_iter).second)->get_symbol();
      }
    }
    
  }
  //---above added for PSG support
  
  SgVariableSymbol *result = NULL;

  // We are searching for the port binding first in the function
  // 'before_end_of_elaboration'
  SgFunctionDefinition *before_end_of_elaboration
    = parent_module->get_before_end_of_elaboration_definition();

  if(before_end_of_elaboration) {
    // Iterating over all statements in the function 'before_end_of_elaboration'
    for(SgStatementPtrList::iterator
        iter  = before_end_of_elaboration->get_body()->get_statements().begin();
        iter != before_end_of_elaboration->get_body()->get_statements().end();
        iter++) {

      // First check if the statement contains a function call
      SgExprStatement *expr_stmt;
      SgFunctionCallExp *func_call;
      if (  (expr_stmt = isSgExprStatement(*iter))
          &&(func_call = isSgFunctionCallExp(expr_stmt->get_expression()))) {
        
        SgVariableSymbol *mapping
          = port_mapping_style1(func_call, port, parent_variable);

        if(mapping) {
          return mapping;
        }

        mapping = port_mapping_style2(func_call, port, parent_variable);
        if(mapping) {
          return mapping;
        }

        mapping
          = port_mapping_style3(func_call, port, parent_module,
              parent_variable);
        if(mapping) {
          return mapping;
        }
      }
    }
  }

  // Second, we are searching for the port binding in the constructor

  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(
        parent_module->get_ast_node()->get_declaration());

  SgFunctionDefinition *ctor_def = ctor->get_definition();
  
  if(!ctor_def) {
    ctor = isSgMemberFunctionDeclaration(ctor->get_definingDeclaration());
    if(ctor == NULL) return NULL;
    ctor_def = ctor->get_definition();
  }

  // Iterating over all statements in the constructor
  for(SgStatementPtrList::iterator
      iter  = ctor_def->get_body()->get_statements().begin();
      iter != ctor_def->get_body()->get_statements().end();
      iter++) {

    // First check if the statement contains a function call
    SgExprStatement *expr_stmt;
    SgFunctionCallExp *func_call;
    if (  (expr_stmt = isSgExprStatement(*iter))
        &&(func_call = isSgFunctionCallExp(expr_stmt->get_expression()))) {

      SgVariableSymbol *mapping
        = port_mapping_style1(func_call, port, parent_variable); 
      if(mapping) {
        return mapping;
      }

      mapping
        = port_mapping_style2(func_call, port, parent_variable); 
      if(mapping) {
        return mapping;
      }

      
      mapping = socket_mapping_style2(func_call, port, parent_variable);
      if(mapping) {
        return mapping;
      }

      mapping
        = port_mapping_style3(func_call, port, parent_module, parent_variable);
      if(mapping) {
        return mapping;
      }
    }
  }

  return NULL;
}


SgVariableSymbol*
risc::get_mapped_symbol_style3(
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable)
{
  /* arguments:
  in parent_module, there is
    parent_variable.port.bind(...)
  or
    parent_variable.port(...)
  */ 

  //this not only works for ports, but also for sockets
  //the port_mapping_style1,2,3 work for both ports 
  //and sockets bindings
  
  //---following added for PSG support
  //if the port mapping for parent_variable.port is already mapped
  //return mapped symbol
  risc::Design* d = risc::get_design();
  for(ModuleInstanceVector::iterator mi_iter = d->loaded_module_instances.begin();
    mi_iter != d->loaded_module_instances.end();
    mi_iter ++)
  {
    if((*mi_iter)->get_ast_node() == parent_variable) 
    {
      // module instance already loaded
      // then we need to find the port
      // and its mapping
      for(std::map<Port*, Object*>::iterator mapping_iter = (*mi_iter)->mapping_.begin();
        mapping_iter!=(*mi_iter)->mapping_.end();
        mapping_iter++)
      {
        
        if(((*mapping_iter).first)->get_symbol()==port) 
          // found the port
          // return the mapping
          return ((*mapping_iter).second)->get_symbol();
      }
    }
    
  }
  //---above added for PSG support
  
  SgVariableSymbol *result = NULL;

  // We are searching for the port binding first in the function
  // 'before_end_of_elaboration'
  SgFunctionDefinition *before_end_of_elaboration
    = parent_module->get_before_end_of_elaboration_definition();

  if(before_end_of_elaboration) {
    // Iterating over all statements in the function 'before_end_of_elaboration'
    for(SgStatementPtrList::iterator
        iter  = before_end_of_elaboration->get_body()->get_statements().begin();
        iter != before_end_of_elaboration->get_body()->get_statements().end();
        iter++) {

      // First check if the statement contains a function call
      SgExprStatement *expr_stmt;
      SgFunctionCallExp *func_call;
      if (  (expr_stmt = isSgExprStatement(*iter))
          &&(func_call = isSgFunctionCallExp(expr_stmt->get_expression()))) {
        
        SgVariableSymbol *mapping
          = port_mapping_style3(func_call, port, parent_module,
              parent_variable);
        if(mapping) {
          return mapping;
        }
      }
    }
  }

  // Second, we are searching for the port binding in the constructor

  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(
        parent_module->get_ast_node()->get_declaration());

  SgFunctionDefinition *ctor_def = ctor->get_definition();
  
  if(!ctor_def) {
    ctor = isSgMemberFunctionDeclaration(ctor->get_definingDeclaration());
    if(ctor == NULL) return NULL;
    ctor_def = ctor->get_definition();
  }

  // Iterating over all statements in the constructor
  for(SgStatementPtrList::iterator
      iter  = ctor_def->get_body()->get_statements().begin();
      iter != ctor_def->get_body()->get_statements().end();
      iter++) {

    // First check if the statement contains a function call
    SgExprStatement *expr_stmt;
    SgFunctionCallExp *func_call;
    if (  (expr_stmt = isSgExprStatement(*iter))
        &&(func_call = isSgFunctionCallExp(expr_stmt->get_expression()))) {

      SgVariableSymbol *mapping
        = port_mapping_style3(func_call, port, parent_module, parent_variable);
      if(mapping) {
        return mapping;
      }
    }
  }

  return NULL;
}


SgVariableSymbol*
risc::get_mapped_module_instance_symbol(
    SgVariableSymbol *socket,
    Module *parent_module,
    SgVariableDefinition *parent_variable)
{
  /* arguments:
  in parent_module, there is
    parent_variable.socket.bind(...)
  or
    parent_variable.socket(...)
  */ 

  SgVariableSymbol *result = NULL;

  // we are searching for the socket binding in the constructor

  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(
        parent_module->get_ast_node()->get_declaration());

  SgFunctionDefinition *ctor_def = ctor->get_definition();
  
  if(!ctor_def) {
    ctor = isSgMemberFunctionDeclaration(ctor->get_definingDeclaration());
    if(ctor == NULL) return NULL;
    ctor_def = ctor->get_definition();
  }

  // Iterating over all statements in the constructor
  for(SgStatementPtrList::iterator
      iter  = ctor_def->get_body()->get_statements().begin();
      iter != ctor_def->get_body()->get_statements().end();
      iter++) {

    // First check if the statement contains a function call
    SgExprStatement *expr_stmt;
    SgFunctionCallExp *func_call;
    if (  (expr_stmt = isSgExprStatement(*iter))
        &&(func_call = isSgFunctionCallExp(expr_stmt->get_expression()))) {

      SgVariableSymbol *mapping
        = module_instance_mapping_style1(func_call, socket, parent_variable);
 
      if(mapping) {
        return mapping;
      }
    }
  }

  return NULL;
}

SgVariableSymbol* risc::port_mapping_style1(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *port,
    SgVariableDefinition *parent_variable)
{
  // TODO: quick hack, I should look for the symbol
  if(SageInterface::get_name(func_call->getAssociatedFunctionDeclaration())
      == "bind") {

    // Please look in a corresponding AST for the documentation
    // given: asdf.jkl.bind(var);
    // var_ref_exp_matching = asdf
    // var_ref = jkl

    SgExpression *function = func_call->get_function();

    SgDotExp *dot_expr = isSgDotExp(function);
    SgCastExp *cast_exp = isSgCastExp(dot_expr->get_lhs_operand());

    SgDotExp *dot_expr2 = NULL;
    SgVarRefExp *var_ref = NULL;
    SgVarRefExp *var_ref_exp_matching = NULL;

    if(cast_exp) {
      dot_expr2 = isSgDotExp(cast_exp->get_operand());

      // work around for sc_out
      if(dot_expr2 == NULL) {

        SgCastExp *cast_exp3 = isSgCastExp(cast_exp->get_operand());

        if(cast_exp3) {
          SgCastExp *cast_exp2 = isSgCastExp(cast_exp3->get_operand());

          SgDotExp *temp_dot_exp_0 = isSgDotExp(cast_exp3->get_operand());

          if(cast_exp2) {
            SgDotExp *temp_dot_exp = isSgDotExp(cast_exp2->get_operand());

            if(temp_dot_exp) {
              dot_expr2 = temp_dot_exp;

            }
          } else if(temp_dot_exp_0) {
              dot_expr2 = temp_dot_exp_0;

          } else {
            assert(0 && "Cannot create port mapping");
          }
        }
      }

      var_ref = isSgVarRefExp(dot_expr2->get_rhs_operand());
      SgArrowExp *arrow_exp = isSgArrowExp(dot_expr2->get_lhs_operand());
      var_ref_exp_matching = isSgVarRefExp(arrow_exp->get_rhs_operand());

    } else {
      // this works for the bind of sc_in
      dot_expr2 = isSgDotExp(dot_expr->get_lhs_operand());

      var_ref_exp_matching
        = isSgVarRefExp(isSgArrowExp(dot_expr2->get_lhs_operand())
            ->get_rhs_operand());

      var_ref = isSgVarRefExp(dot_expr2->get_rhs_operand());
    }

    if(var_ref->get_symbol() == port) {

      // Here we are checking, if the parent variable is matching the
      // binding variable.
      if(var_ref_exp_matching->get_symbol() ==
          parent_variable->get_vardefn()->
            search_for_symbol_from_symbol_table()) {


        // is binded to this one
        std::vector<SgNode*> binded_element
          = NodeQuery::querySubTree(func_call->get_args(), V_SgVarRefExp);

        return isSgVarRefExp(binded_element.at(0))->get_symbol();
      }
    }
  }

  return NULL;
}

SgVariableSymbol* risc::port_mapping_style2(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *port,
    SgVariableDefinition *parent_variable)
{
  //for binding like a.b(c)
  SgDotExp *dot = isSgDotExp(func_call->get_function());
  if(dot) {

    SgMemberFunctionRefExp *mem_func_ref
      = isSgMemberFunctionRefExp(dot->get_rhs_operand());

    if(mem_func_ref) {
      
      if(SageInterface::get_name(mem_func_ref)
          != "member_function_ref_operator()") {
        return NULL;
      }
    } else {
      return NULL;
    }

    SgCastExp *cast_exp = isSgCastExp(dot->get_lhs_operand());
    if(cast_exp) {

      SgCastExp *cast_exp2 = isSgCastExp(cast_exp->get_operand());
      if(cast_exp2) {

        SgDotExp *dot2 = isSgDotExp(cast_exp2->get_operand());
        if(dot2) {

          SgArrowExp *arrow_exp = isSgArrowExp(dot2->get_lhs_operand());
          if(arrow_exp) {

            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
            if(var_ref) {
              if(var_ref->get_symbol()
                  != parent_variable->get_declaration()->get_decl_item(NULL)
                        ->get_symbol_from_symbol_table()) {
                return NULL;
              }
            } else {
              return NULL;
            }
          }

          SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
          if(var_ref) {
            if(var_ref->get_symbol() != port) {
              return NULL;
            }
          } else {
            return NULL;
          }
        }
      } else {

        SgDotExp *dot2 = isSgDotExp(cast_exp->get_operand());

        if(dot2) {

          SgArrowExp *arrow_exp = isSgArrowExp(dot2->get_lhs_operand());

          if(arrow_exp) {

            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());

            if(var_ref) {
              if(var_ref->get_symbol()
                  != parent_variable->get_declaration()->get_decl_item(NULL)
                        ->get_symbol_from_symbol_table()) {
                return NULL;
              }
            } else {
              return NULL;
            }
          }

          SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
          if(var_ref) {
            if(var_ref->get_symbol() != port) {
              return NULL;
            }
          } else {
            return NULL;
          }
        }
      }
    } else {
      return NULL;
    }
  }

  SgExprListExp *expr = func_call->get_args();
  if(expr) {

    SgExpressionPtrList &list = expr->get_expressions();

    if(list.size() != 1) {
      return NULL;
    }

    SgCastExp *cast = isSgCastExp(list.at(0));

    if(cast) {
      SgArrowExp *arrow = isSgArrowExp(cast->get_operand());
      if(arrow) {
        SgVarRefExp *var_ref = isSgVarRefExp(arrow->get_rhs_operand());
        if(var_ref) {
          return var_ref->get_symbol();
        } else {
          return NULL;
        }
      }
    } else {
      return NULL;
    }
  }

  return NULL;
}

SgVariableSymbol* risc::port_mapping_style3(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable)
{

  ModuleInstance *sub_module_instance
    = parent_module->find_module(parent_variable);

  Module *sub_module = sub_module_instance->get_definition();

  int port_id = 0;
  for(SgDeclarationStatementPtrList::iterator
      iter  = sub_module->get_ast_node()->get_members().begin();
      iter != sub_module->get_ast_node()->get_members().end();
      iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);

    if(var_decl == NULL) {
      continue;
    }

    SgVariableDefinition *var_def = var_decl->get_definition();
    SgType* var_type = var_def->get_type();

    if(risc::tools::is_data_structure(var_type)) {

      SgClassType *class_type = isSgClassType(var_type);
      SgClassDeclaration *class_decl
        = isSgClassDeclaration(class_type->get_declaration());

      if(class_decl->isForward()) {
        class_decl
          = isSgClassDeclaration(class_decl->get_definingDeclaration());
      }

      SgClassDefinition *class_def = class_decl->get_definition();

      if(tools::is_derived_from_sc_port_base(class_def) ) {

        if(port == var_def->get_vardefn()
                      ->search_for_symbol_from_symbol_table() ) {

          break;
        }
        port_id++;
      }
    }
  }

  SgDotExp *dot = isSgDotExp(func_call->get_function());
  if(dot) {

    SgMemberFunctionRefExp *mem_func_ref
      = isSgMemberFunctionRefExp(dot->get_rhs_operand());

    if(mem_func_ref) {
      if(SageInterface::get_name(mem_func_ref)
          != "member_function_ref_operator()") {
        return NULL;
      }
    } else {
      return NULL;
    }

    SgCastExp *cast = isSgCastExp(dot->get_lhs_operand());

    if(cast) {
      SgArrowExp *arrow = isSgArrowExp(cast->get_operand());

      if(arrow) {

        SgVarRefExp *var_ref = isSgVarRefExp(arrow->get_rhs_operand());

        if(var_ref) {

          if(var_ref->get_symbol()
              != parent_variable->get_declaration()->get_decl_item(NULL)
                     ->get_symbol_from_symbol_table()) {
            return NULL;
          }
        } else {
          return NULL;
        }
      } else {
        return NULL;
      }
    } else {
      return NULL;
    }
  } else {
    return NULL;
  }

  SgExprListExp *expr = func_call->get_args();
  if(expr) {

    SgExpressionPtrList &list = expr->get_expressions();

    SgConstructorInitializer *ctor_init
      = isSgConstructorInitializer(list.at(port_id));
    if(ctor_init) {
      SgExprListExp *list = ctor_init->get_args();

      if(list) {

        SgCastExp *cast = isSgCastExp(*(list->get_expressions().begin()));

        if(cast) {

          SgArrowExp *arrow = isSgArrowExp(cast->get_operand());

          if(arrow) {
            SgVarRefExp *var_ref = isSgVarRefExp(arrow->get_rhs_operand());
            if(var_ref) {
              return var_ref->get_symbol();
            } else {
              return NULL;
            }

          } else {
            return NULL;
          }

        } else {
          return NULL;
        }

      } else {
        return NULL;
      }

    } else {
      return NULL;
    }
  }

  return NULL;
}


SgVariableSymbol* risc::socket_mapping_style2(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *socket,
    SgVariableDefinition *parent_variable)
{

  
  //this is a.b(c.d)
  SgDotExp *dot = isSgDotExp(func_call->get_function());
  if(dot) {

    SgMemberFunctionRefExp *mem_func_ref
      = isSgMemberFunctionRefExp(dot->get_rhs_operand());

    if(mem_func_ref) {
      if(SageInterface::get_name(mem_func_ref)
          != "member_function_ref_operator()") {
        return NULL;
      }
      
    } else {
      return NULL;
    }
    SgCastExp *cast_exp = isSgCastExp(dot->get_lhs_operand());
    if(cast_exp) {

      SgCastExp *cast_exp2 = isSgCastExp(cast_exp->get_operand());
      if(cast_exp2) {

        SgDotExp *dot2 = isSgDotExp(cast_exp2->get_operand());
        if(dot2) {

          SgArrowExp *arrow_exp = isSgArrowExp(dot2->get_lhs_operand());
          if(arrow_exp) {

            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
            if(var_ref) {
              if(var_ref->get_symbol()
                  != parent_variable->get_declaration()->get_decl_item(NULL)
                        ->get_symbol_from_symbol_table()) {
                return NULL;
              }
            } else {
              return NULL;
            }
          }

          SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
          if(var_ref) {
            if(var_ref->get_symbol() != socket) {
              return NULL;
            }
          } else {
            return NULL;
          }
        }
      } else {

        SgDotExp *dot2 = isSgDotExp(cast_exp->get_operand());

        if(dot2) {

          SgArrowExp *arrow_exp = isSgArrowExp(dot2->get_lhs_operand());

          if(arrow_exp) {

            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());

            if(var_ref) {
              if(var_ref->get_symbol()
                  != parent_variable->get_declaration()->get_decl_item(NULL)
                        ->get_symbol_from_symbol_table()) {
                return NULL;
              }
            } else {
              return NULL;
            }
          }

          SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
          if(var_ref) {
            if(var_ref->get_symbol() != socket) {
              return NULL;
            }
          } else {
            return NULL;
          }
        }
      }
    } else {
      return NULL;
    }
  }
  SgExprListExp *expr = func_call->get_args();
  if(expr) {

    SgExpressionPtrList &list = expr->get_expressions();

    if(list.size() != 1) {
      return NULL;
    }

    SgCastExp *cast = isSgCastExp(list.at(0));

    if(cast) {

      //get the mapped socket d
      SgCastExp *cast2 = isSgCastExp(cast->get_operand());
      if(cast2){
        SgCastExp *cast3 = isSgCastExp(cast2->get_operand());
        if(cast3){
          SgDotExp* dot2 = isSgDotExp(cast3->get_operand());
          if(dot2){
            SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
            if(var_ref) {
              return var_ref->get_symbol();
            } else {
              return NULL;
            }
          }
        }
      }
      //done
    } else {
      return NULL;
    }
  }

  return NULL;
}


SgVariableSymbol* risc::module_instance_mapping_style1(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *socket,
    SgVariableDefinition *parent_variable)
{
  //this is parent_variable.socket(c.d)
  //and returns symbol of c
  SgDotExp *dot = isSgDotExp(func_call->get_function());
  if(dot) {

    SgMemberFunctionRefExp *mem_func_ref
      = isSgMemberFunctionRefExp(dot->get_rhs_operand());

    if(mem_func_ref) {
      if(SageInterface::get_name(mem_func_ref)
          != "member_function_ref_operator()") {
        return NULL;
      }
      
    } else {
      return NULL;
    }
    SgCastExp *cast_exp = isSgCastExp(dot->get_lhs_operand());
    if(cast_exp) {

      SgCastExp *cast_exp2 = isSgCastExp(cast_exp->get_operand());
      if(cast_exp2) {

        SgDotExp *dot2 = isSgDotExp(cast_exp2->get_operand());
        if(dot2) {

          SgArrowExp *arrow_exp = isSgArrowExp(dot2->get_lhs_operand());
          if(arrow_exp) {

            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
            if(var_ref) {
              if(var_ref->get_symbol()
                  != parent_variable->get_declaration()->get_decl_item(NULL)
                        ->get_symbol_from_symbol_table()) {
                return NULL;
              }
            } else {
              return NULL;
            }
          }

          SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
          if(var_ref) {
            if(var_ref->get_symbol() != socket) {
              return NULL;
            }
          } else {
            return NULL;
          }
        }
      } else {

        SgDotExp *dot2 = isSgDotExp(cast_exp->get_operand());

        if(dot2) {

          SgArrowExp *arrow_exp = isSgArrowExp(dot2->get_lhs_operand());

          if(arrow_exp) {

            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());

            if(var_ref) {
              if(var_ref->get_symbol()
                  != parent_variable->get_declaration()->get_decl_item(NULL)
                        ->get_symbol_from_symbol_table()) {
                return NULL;
              }
            } else {
              return NULL;
            }
          }

          SgVarRefExp *var_ref = isSgVarRefExp(dot2->get_rhs_operand());
          if(var_ref) {
            if(var_ref->get_symbol() != socket) {
              return NULL;
            }
          } else {
            return NULL;
          }
        }
      }
    } else {
      return NULL;
    }
  }
  SgExprListExp *expr = func_call->get_args();
  if(expr) {

    SgExpressionPtrList &list = expr->get_expressions();

    if(list.size() != 1) {
      return NULL;
    }

    SgCastExp *cast = isSgCastExp(list.at(0));

    if(cast) {

      
      SgCastExp *cast2 = isSgCastExp(cast->get_operand());
      if(cast2){
        SgCastExp *cast3 = isSgCastExp(cast2->get_operand());
        if(cast3){
          SgDotExp* dot2 = isSgDotExp(cast3->get_operand());
          if(dot2){

            SgArrowExp* arrow1 = isSgArrowExp(dot2->get_lhs_operand());

            if(arrow1)
            {
              SgVarRefExp *var_ref = isSgVarRefExp(arrow1->get_rhs_operand());
              if(var_ref) {
                return var_ref->get_symbol();
              } else {
                return NULL;
              }
            }
          }
        }
      }
      //done
    } else {
      return NULL;
    }
  }

  return NULL;
}

risc::InstanceTree
risc::get_mapped_channel_instance(InstanceTree port)
{

  if(PSG_MODE) return get_mapped_channel_instance_for_psg(port);
  Object* port_obj = port.get_object();
  SgVariableSymbol *port_symbol = port.get_object()->get_symbol();
  
  // going one step up in the hierarchy
  port.path_.pop_back();
  // getting the mapped element in the parent class

  SgVariableDefinition *mapped_object_def = NULL;
  for(std::map<risc::Port*, risc::Object*>::iterator 
    it  = (dynamic_cast<risc::ModuleInstance*>(port.get_object()))->mapping_.begin();
    it != (dynamic_cast<risc::ModuleInstance*>(port.get_object()))->mapping_.end();
    it ++)
  {
    if(it->first->get_ast_node() == port_obj->get_ast_node())
    {
      mapped_object_def = it->second->get_ast_node();
    }
  }
  
  if(!mapped_object_def)
  {
    std::cout << "Error: Unsolved mapping."
              << std::endl;
    std::cout << "       Port "
              << port_obj->get_name()
              << " of instance "
              << port.get_object()->get_name()
              << " in "
              << port.get_scope()->get_name()
              << " is not bound."
              << std::endl;
    exit(1);
  }
  // here we are checking the connection is a port
  Port *intermediate_port = port.get_scope()->find_port(
      isSgVariableDefinition(mapped_object_def));

  // here we are going in the hierarchy up until the is no longer a port to
  // port mapping
  while(intermediate_port) {

    //port_symbol = intermediate_port->get_symbol();

    // going one step up in the hierarchy
    port.path_.pop_back();
    for(std::map<risc::Port*, risc::Object*>::iterator 
      it  = (dynamic_cast<risc::ModuleInstance*>(port.get_object()))->mapping_.begin();
      it != (dynamic_cast<risc::ModuleInstance*>(port.get_object()))->mapping_.end();
      it ++)
    {
      if(it->first->get_symbol() == intermediate_port->get_symbol())
      {
        
        mapped_object_def = it->second->get_ast_node();
      }
    }
    //break;
    intermediate_port = port.get_scope()->find_port(
      isSgVariableDefinition(mapped_object_def));
  }

  // Finally, we have to create the instance tree for the mapped channel
  // In the following, we are sorting out if it is a primitive or a
  // hierarchical channel
  // SgVariableDefinition *channel_var_def =
  //   isSgVariableDefinition(mapped_object->get_declaration()->get_definition());
  SgVariableDefinition *channel_var_def =
    isSgVariableDefinition(mapped_object_def);
  Instance *mapped_channel
    = port.get_scope()->find_prim_channel(channel_var_def);

  if(mapped_channel) {

    port.path_.back().second = mapped_channel;
  } else {

    port.path_.back().second
      = port.get_scope()->find_hier_channel(channel_var_def);
  }
  return port;
}

// OLD CODE, NOT CORRECT WHEN THERE IS SOCKET MAPPING
// KEEPS HERE ONLY FOR PSG
risc::InstanceTree
risc::get_mapped_channel_instance_for_psg(InstanceTree port)
{
  Object* port_obj = port.get_object();
  SgVariableSymbol *port_symbol = port.get_object()->get_symbol();
  
  // going one step up in the hierarchy
  port.path_.pop_back();
  // getting the mapped element in the parent class
  SgVariableSymbol *mapped_object = get_mapped_symbol(
      port_symbol,
      dynamic_cast<Module*>(port.get_scope()),
      port.get_object()->get_ast_node());

  if(!mapped_object)
  {
    std::cout << "Error: Unsolved mapping."
              << std::endl;
    std::cout << "       Port "
              << port_obj->get_name()
              << " of instance "
              << port.get_object()->get_name()
              << " in "
              << port.get_scope()->get_name()
              << " is not bound."
              << std::endl;
    exit(1);
  }
  // here we are checking the connection is a port
  Port *intermediate_port = port.get_scope()->find_port(
      isSgVariableDefinition(mapped_object->get_declaration()
        ->get_definition()));

  // here we are going in the hierarchy up until the is no longer a port to
  // port mapping
  while(intermediate_port) {

    port_symbol = intermediate_port->get_symbol();

    // going one step up in the hierarchy
    port.path_.pop_back();

    // getting the mapped element in the parent class
    mapped_object = get_mapped_symbol(
                        port_symbol,
                        dynamic_cast<Module*>(port.get_scope()),
                        port.get_object()->get_ast_node());

    // here we are checking the connection is a port
    intermediate_port = port.get_scope()->find_port(
      isSgVariableDefinition(mapped_object->get_declaration()
        ->get_definition()));
  }

  // Finally, we have to create the instance tree for the mapped channel
  // In the following, we are sorting out if it is a primitive or a
  // hierarchical channel
  SgVariableDefinition *channel_var_def =
    isSgVariableDefinition(mapped_object->get_declaration()->get_definition());

  Instance *mapped_channel
    = port.get_scope()->find_prim_channel(channel_var_def);

  if(mapped_channel) {

    port.path_.back().second = mapped_channel;
  } else {

    port.path_.back().second
      = port.get_scope()->find_hier_channel(channel_var_def);
  }

  return port;
}


risc::InstanceTree
risc::get_mapped_socket_instance_tree(InstanceTree instance_tree_1)
{
//  std::cout << "\nentered get_mapped_socket_instance_tree, print instance_tree_1" << std::endl;
  //instance_tree_1.print_instance_tree();
  //module_instance_1.socket_1(module_instance_2.socket_2)
  //returns the instance tree of ..->parent_module_instance->module_instance_2->socket_2
  
  //initially, instance_tree_1 is ..->parent_module_instance->module_instance_1->socket_1
  Socket* socket_1 = dynamic_cast<Socket*>(instance_tree_1.get_object());
  //std::cout << "socket_1 type is " << socket_1->type_ << std::endl;
  
  instance_tree_1.path_.pop_back();
  //after the above line, instance_tree_1 is ..->parent_module_instance->module_instance_1
  //std::cout << "\nafter popback, print instance_tree_1" << std::endl;
  //instance_tree_1.print_instance_tree();
  
  ModuleInstance* module_instance_1 
    = dynamic_cast<ModuleInstance*>(instance_tree_1.get_object());
  //std::cout << "module_instance_1 name is " << module_instance_1->get_name() << std::endl;
  //if socket1 is not mapped, then return an empty InstanceTree
  if( module_instance_1->socket_mapping_.count( socket_1) == 0){
    risc::InstanceTree tmp;
    return tmp;
  }

  Socket* socket_2 
    = module_instance_1->socket_mapping_[socket_1];
  //std::cout << "socket_2 name is " << socket_2->get_name() << std::endl;
  SgVariableSymbol *module_instance_2_symbol = 
    get_mapped_module_instance_symbol(
      socket_1->get_symbol(),
      dynamic_cast<Module*>(instance_tree_1.get_scope()),
      module_instance_1->get_ast_node()
    );
  //std::cout << "module_instance_2_symbol name is " << module_instance_2_symbol->get_name() << std::endl;
  if(module_instance_2_symbol == NULL)
  {
    //meaning the socket mapping is like:
    //module_instance1.socket1.bind(socket2)
    //and socket2 is only an intermediate socket
    instance_tree_1.path_.pop_back();
    
    InstanceTree::Element sub_path;
    sub_path.first = dynamic_cast<ModuleInstance*>(instance_tree_1.get_object())->get_definition();
    sub_path.second = socket_2;
    instance_tree_1.path_.push_back(sub_path);

    return get_mapped_socket_instance_tree(instance_tree_1);
  }
  else
  {
    ModuleInstance *module_instance_2
      = instance_tree_1.get_scope()->find_module(
          isSgVariableDefinition(module_instance_2_symbol->get_declaration()
                                  ->get_definition()));
                                  
    //std::cout << "module_instance_2 name is " << module_instance_2->get_name() << std::endl;

    instance_tree_1.path_.back().second = module_instance_2;
    //after the above line, instance_tree_1 is ..->parent_module_instance->module_instance_2

    InstanceTree::Element sub_path;

    sub_path.first = module_instance_2->get_definition();
    sub_path.second = socket_2;

    instance_tree_1.path_.push_back(sub_path);
    //after the above line, 
    //instance_tree_1 is ..->parent_module_instance->module_instance_2->socket_2
    
    return instance_tree_1;
  }
  
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
