#ifndef SEGMENT_GRAPH_H_INCLUDED_
#define SEGMENT_GRAPH_H_INCLUDED_

#include<iostream>
#include<map>
#include<set>
#include<utility>

#include "boost/graph/adjacency_list.hpp"
#include "boost/graph/graphviz.hpp"
#include "boost/algorithm/string.hpp"
#include "../internal_representation/module.h"

#include "rose.h"

#include "edge.h"
#include "function_annotation.h"
#include "port_call_path.h"
#include "segment.h"


#define DBG_SHOW(name,id) std::cout<<"\n\ndebug id: "<< id<<"  "<<#name<<" :  \n"<<name->unparseToString()<<std::endl;
namespace risc {

class Function;

namespace sg {

class SegmentGraph;

typedef boost::adjacency_list<boost::listS, // list (not vector!) of vertices (RD)
        boost::listS, boost::bidirectionalS, Segment, Edge,
        boost::property<boost::vertex_index_t, int> > Graph;
typedef boost::graph_traits<Graph>::vertex_descriptor VertexDescriptor; // ends up as 'void*' (RD)
typedef boost::graph_traits<Graph>::vertex_iterator VertexIterator;     // ends up as 'void*' (RD)
typedef boost::graph_traits<Graph>::edge_descriptor EdgeDescriptor;

class FunctionAnnotationAttributes;
class CachedFunctionAstAttributes;

/**
 *  \class SegmentCmp
 *  \brief Function object (functor!?) for SegmentSet (12/15/18, RD)
 *
 */
struct SegmentCmp {
  bool operator()(                  // comparison of two VertexDescriptors
        const VertexDescriptor lhs,
        const VertexDescriptor rhs) const
  {
//  return lhs < rhs; // this would be the default behavior, comparing 2 pointers;
                      // because that's non-deterministic over multiple program runs,
                      // we compare instead the segment IDs (12/15/18, RD)

    Graph *FakeGraphHandle = NULL;  // we only need this type, not the value!
                                    // (the VertexDescriptors are 'void*' pointers,
                                    //  allowing direct access to the vertices
                                    //  without knowing the specific Graph object;
                                    //  this works only for a _list_ of vertices)

//  printf("DEBUG: SegmentCmp(lhs=%p, rhs=%p) lhs.id=%d, rhs.id=%d, lhs.node_id=%d, rhs.node_id=%d\n",
//          lhs, rhs,
//          (*FakeGraphHandle)[lhs].id_, (*FakeGraphHandle)[rhs].id_,
//          (*FakeGraphHandle)[lhs].node_id_, (*FakeGraphHandle)[rhs].node_id_);
//  assert((*FakeGraphHandle)[lhs].id_ == (*FakeGraphHandle)[rhs].id_ ?
//              lhs == rhs : true); // this fails in PSG cases
    assert((*FakeGraphHandle)[lhs].node_id_ == (*FakeGraphHandle)[rhs].node_id_ ?
                lhs == rhs : true); // this holds true even for PSG segments

    return (*FakeGraphHandle)[lhs].node_id_ < (*FakeGraphHandle)[rhs].node_id_;
  }
};

/**
 *  \class SegmentGraph
 *  \brief This class represents a segment graph for a process
 *
 */
class SegmentGraph {

public:

  typedef std::set<SgFunctionDefinition*> FunctionCallBoundarySet;
  typedef std::set<VariantT> CCxxKeywordsBoundarySet;

  SegmentGraph();  
  SegmentGraph(FunctionCallBoundarySet function_boundaries);
  SegmentGraph(CCxxKeywordsBoundarySet keyword_boundaries);
  SegmentGraph(FunctionCallBoundarySet function_boundaries,
      CCxxKeywordsBoundarySet keyword_boundaries);

  typedef std::set<VertexDescriptor,              // RD's new one  (deterministic order by segment IDs)
                    SegmentCmp> SegmentSet;       // (comparison performed by SegmentCmp above)

  /**
   * \fn        get_wait_and_notifying_events
   * \brief     set the wait_events_ and notify_events_ for psg nodes
   *            under psg mode
   */
  void get_wait_and_notifying_events();

  /**
   * \fn        find_module_segments
   * \brief     for loading PSG
   *            It adds simulation threads to module_definitions,
   *            and sets the corresponding starting_id_ of the thread.
   */
  void find_module_segments();

  /**
   * \fn        print_psg_file(std::string file_name)
   * \brief     store the psg as .psg file on disk
   *
   * \param     file_name   The name of the file
   */
  void print_psg_file(std::string file_name);

  /**
   * \fn        read_psg_from_file(std::string)
   * \brief     load the .psg file from disk
   *            reconstructs nodes and connections
   *
   * \param     file_name   The name of the file
   */
  void read_psg_from_file(std::string file_name);


  /**
   * \fn        replace_partial_function_call_nodes
   * \brief     for each partial function call node,
   *            calls integrate_psg_into_sg to replace the node
   *            with corresponding sub psg
   *
   */
  void replace_partial_function_call_nodes();

  /**
   * \fn        integrate_psg_into_sg(VertexDescriptor, SgSymbol *)
   * \brief     replace a partial function call node with its
   *            psg. If the partial function call is a port call,
   *            the integrated psg will be cloned, and set pcp
   *
   * \param     func_call_vertex_id   vertex descriptor of the 
   *                                  function call node
   * \param     func_sym    Symbol of the function call, used to
   *                        locate the entrance and exit nodes of
   *                        the sub psg
   */
  void integrate_psg_into_sg(VertexDescriptor func_call_vertex_id, SgSymbol * func_sym);

  /**
   * \fn        remove_dangling_segments
   * \brief     after integration of all the psgs, the segments that
   *            does not belong to simulation processes are removed
   */
  void remove_dangling_segments();
  
  /**
   * \brief we need to instrument set_upcoming_segment_id(var_array)
   *        before port calls
   *        var_array is defined in the top level
   *        channel_seg_id_array_names_ stores the names of all var_array.
   *        the names are passed to instrumentor to instrument the arrays
   *        in the top file.
   *        each name was stored in psg files and loaded back
   */
  std::vector<std::string> channel_seg_id_array_names_;

  /**
   * \brief stores the partial function call nodes that hold port calls.
   *        are used in PSG mode.
   */
  std::set<VertexDescriptor> port_call_seg_vertexs_;

  /**
   * \brief pairs of VertexDescriptor of partial function call node
   *        and the function's symbol.
   *        used for replacing partial function call nodes with its 
   *        subpsg
   */
  std::vector<std::pair<VertexDescriptor, SgSymbol *> > 
    vertexID_symbol_pairs_of_partialFuncCallNodes_;  

  /**
   * \brief mapping of function symbol and its entrance segment vertexes,
   *        reconstructed after loading
   */
  std::map< SgSymbol * 	, std::vector<VertexDescriptor> >
    func_symbols_and_entrance_segVertex_;

  /**
   * \brief mapping of function symbol and its exit segment vertexes
   *        reconstructed after loading
   */
  std::map< SgSymbol * 	, std::vector<VertexDescriptor> >
    func_symbols_and_exit_segVertex_;

  /**
   * \brief for redaction of psg, we need to store the total number 
   *        of segment nodes. The value is set in the main function
   */
  int total_segment_nodes_;
   
  /**
   * \fn       clean_graph
   * \brief    This function deletes all duplicated nodes in each vertex
   *           of the graph
   */
  void clean_graph();


  /**
   * \fn       print_graph(std::string filename)
   * \brief    This function prints the graph in a dot file.
   *           The name of the file is user defined by argument
   * \param    filename   The name of the file
   */
  void print_graph(std::string filename);


  /**
   * \fn       print_graph_read_write_access(std::string filename)
   * \brief    This function prints all variables which have read or written
   *           in this segment.
   * \param    filename   The name of the file
   */
  void print_graph_read_write_access(std::string filename);


  /**
   * \fn       add_function(
                Thread *originating_thread,
                Function *function,
                bool duplicate_segments = false,
                bool is_simulation_process = true,
                std::string module_name = "")
   * \brief    This function analyzes the given function and creates a
   *           segment graph for it. The graph will be automatically inserted
   *           in the existing graph. The flag duplicate_segments will be
   *           used to generate an instance segement graph.
   *           is_simulation_process: if set to true, meaning the function is a 
   *           simulation process, and it needs a segment id
   */
  void add_function(
        Thread *originating_thread,
        Function *function,
        bool duplicate_segments = false,
        bool is_simulation_process = true,
        std::string module_name = "");

  //DM 05/20/2019 same as add_function with extra functionality for sc_method
  /**
   * \fn       add_function_method(
                Thread *originating_thread,
                Function *function,
                bool duplicate_segments = false,
                bool is_simulation_process = true,
                std::string module_name = "")
   * \brief    This function analyzes the given function and creates a
   *           segment graph for it. The graph will be automatically inserted
   *           in the existing graph. The flag duplicate_segments will be
   *           used to generate an instance segement graph.
   *           is_simulation_process: if set to true, meaning the function is a 
   *           simulation process, and it needs a segment id
   */
  void add_function_method(
        Thread *originating_thread,
        Function *function,
        bool duplicate_segments = false,
        bool is_simulation_process = true,
        std::string module_name = "");

  /**
   * \fn       read_write_analysis_of_segments
   * \brief    The function determines which variables are read and written in
   *           the individual segments
   */
  void read_write_analysis_of_segments();

  /**
   * \fn       set_all_segments_to_untouched
   * \brief    This function sets the status of the flag
   *           helper_for_graph_algorithms for all nodes to 'Untouched'
   */
  void set_all_segments_to_untouched();

private:

  /**
   * \brief   determine if the compiler should perform
   *          static analysis or dynamic analysis
   */
  bool static_analysis_;

  /**
   * \fn        handle_recursive_calls
   * \brief     This function integrates the recursive function calls
   */
  void handle_recursive_calls();


  /**
   * \fn        separate_variable_declaration_and_initializer_on_demand(
           SgBasicBlock *&bb, bool duplicate_segments, PortCallPath pcp)
   * \brief     This function checks if a direct or indirect boundary call is
   *            in the initializer expression of the variable declaration.
   *            If yes, the declaration and initialization will be separated.
   *            Example:
   *            Input:
   *                int x = y + foo(); // foo() has a indirect boundary
   *            Output:
   *                int x;
   *                x = y + foo();
   */
  void separate_variable_declaration_and_initializer_on_demand(
           SgBasicBlock *&bb, bool duplicate_segments, PortCallPath pcp);


  /**
   * \fn        decompose_expression_with_boundary_calls(SgExpression *expr,
           SgBasicBlock *&bb, bool duplicate_segments, PortCallPath pcp,
           int stmt_type = 0, SgStatement* parent = NULL,
           SgBasicBlock* body = NULL)
   * \brief     This function decomposes an expression into multiple
   *            statements. Functions which call directly or indirectly a
   *            boundary will be separated and have temporally variables.
   *            Example:
   *            Input:
   *                x = y + foo(); // foo() has a indirect boundary
   *            Output:
   *                int temp_foo = foo();
   *                int x = y + temp_foo();
   */
  
  void decompose_expression_with_boundary_calls(SgExpression *expr,
           SgBasicBlock *&bb, bool duplicate_segments, PortCallPath pcp,
           int stmt_type = 0, SgStatement* parent = NULL,
           SgBasicBlock* body = NULL);

  /**
   * \fn       create_temp_variable_for_expression(SgExpression *expr,
           SgBasicBlock *&bb,
           int stmt_type = 0, SgStatement* parent = NULL,
           SgBasicBlock* body = NULL)
   * \brief    This function creates a variable declaration of the type of the
   *           given expression. The variable will be inserted before the
   *           associated expression statement of the given expression.
   */
  void create_temp_variable_for_expression(SgExpression *expr,
           SgBasicBlock *&bb,
           int stmt_type = 0, SgStatement* parent = NULL,
           SgBasicBlock* body = NULL);

  /**
   * \fn       generate_unique_name(SgNode *node)
   * \brief    Generates an unique name for a varible based on the given node.
   *
   */
  std::string generate_unique_name(SgNode *node);

  /**
   * \fn       generate_unique_name(std::string)
   * \brief    Generates an unique name for a varible based on the given string.
   *
   */
  std::string generate_unique_name(std::string);


  /**
   * \fn         is_boundary_stmt(SgStatement const * const current_stmt)
   * \brief      This function checks if a given statement is a potential
   *             boundary statement
   * \param      current_stmt The statement which will be checked
   */
  bool is_boundary_stmt(SgStatement const * const current_stmt);


  /**
   * \fn       has_function_call_with_boundary(SgExpression *expr,
                         bool duplicate_segments, PortCallPath pcp)
   * \brief    This function checks if in the given expression a function
   *           call occurs which calls a functon with a boundary statements.
   */
  SgFunctionCallExp* has_function_call_with_boundary(SgExpression *expr,
                         bool duplicate_segments, PortCallPath pcp);


  /**
   * \fn       should_decompose_function(SgExpression *expr, bool duplicate_segments,
      PortCallPath pcp, bool is_condition= false)
   * \brief    This function determines if a expression should be decomposed.
   *           This is the case if expression has the following structure:
   *           var = somepointer->func();
   *           var = somevariable.func();
   *           var = func();
   *           and var is a global, member, or static variable.
   */
  bool should_decompose_function(SgExpression *expr, bool duplicate_segments,
      PortCallPath pcp, bool is_condition= false);


  /**
   * \fn       contains_function_call_expression(SgExpression *expr)
   * \brief    In this function we look for expressions which contain a function
   *           call in the following style:
   *           foo();
   *           this->foo();
   *           var = foo();
   *           var = this->foo();
   *           The return value will be the function declaration of the called
   *           function
   */
  SgFunctionDeclaration* contains_function_call_expression(SgExpression *expr);


  /**
   * \fn       build_graph(
      SgStatement *current_stmt,
      SegmentSet current_segments,
      SegmentSet &break_segments,
      SegmentSet &continue_segments,
      bool duplicate_segments,
      PortCallPath pcp)
   * \brief    This function creates the segment graph
   *
   * \param    current_stmt        The first statement of the sequence which
   *                               will be analyzed
   * \param    current_segments    The beginning segments
   * \param    break_segments    The break segments
   * \param    continue_segments    The continue segments
   * \param    duplicate_segments    clone or link
   * \param    pcp    The pcp information
   */
  SegmentSet build_graph(
      SgStatement *current_stmt,
      SegmentSet current_segments,
      SegmentSet &break_segments,
      SegmentSet &continue_segments,
      bool duplicate_segments,
      PortCallPath pcp);


  /**
   * \fn       add_expression_to_segment(SegmentSet segments, SgNode *expr,
      PortCallPath pcp)
   * \brief    This function adds to each segment the given expression
   *
   * \param    segments   These segments which will contain the given
   *                      expression
   * \param    expr       The expressions which will added to all segments
   * \param    pcp        add the pcp information
   */
  void add_expression_to_segment(SegmentSet segments, SgNode *expr,
      PortCallPath pcp);


  /**
   * \fn       build_segment_graph_for_function(
      SgFunctionDefinition *func_def,
      bool duplicate_segments,
      PortCallPath pcp,
      bool is_registered_b_transport = false)
   * \brief    This function generates a segment graph for a function
   */
  CachedFunctionAstAttributes*
  build_segment_graph_for_function(
      SgFunctionDefinition *func_def,
      bool duplicate_segments,
      PortCallPath pcp,
      bool is_registered_b_transport = false);


  /**
   * \fn       insert_loop_edges(VertexDescriptor &loop_vertex,
           SegmentSet &current_segments,
           SegmentSet &continue_segments,
           SegmentSet &leaf_segments_of_loop)
   * \brief    This is a helper function for loop handling
   *
   */
  void insert_loop_edges(VertexDescriptor &loop_vertex,
           SegmentSet &current_segments,
           SegmentSet &continue_segments,
           SegmentSet &leaf_segments_of_loop);

  /**
   * \fn    get_function_call_annotation(
      SgFunctionCallExp *func_call_exp)
   * \brief This function takes a function call expression and checks if there
   *        is any function annotation at the declaration. If the function
   *        declaration has no annotation, there will be a check if a annotation
   *        must be created
   *
   *        FA: Function annoation
   *        CF: Cached function
   *
   *                           comment        no comment
   *        definition      FA (inlining)    CF (linking)
   *        no defiition    FA (inlining)    FA (nothing,
   *                                         default: no-wait conflict-free)
   */
  FunctionAnnotationAttributes* get_function_call_annotation(
      SgFunctionCallExp *func_call_exp);

  /**
   * \fn    create_cached_function_attribute_for_annotated_function(
      SgNode* boundary_node,
      WAIT_CONSTRUCT wait_type,
      SegmentSet incoming_segments,
      bool is_conflict_free,
      PortCallPath pcp)
   * \brief This function creates a cached function annotation information
   *        for based on the cached function information
   */
  CachedFunctionAstAttributes*
  create_cached_function_attribute_for_annotated_function(
      SgNode* boundary_node,
      WAIT_CONSTRUCT wait_type,
      SegmentSet incoming_segments,
      bool is_conflict_free,
      PortCallPath pcp);

  /**
   * \fn    add_mapped_symbol_to_reference(SgFunctionCallExp *func_call,
           SgFunctionDefinition *func_def)
   * \brief Let's take a look at this example:
   *        void foo(int &ref, int value, in &ref2);
   *        foo(var, some_var, other_var);
   *        This function adds the symbol of "var" to the attribute "references"
   *        of the symbol "ref".
   *        The same happens for "ref2" and "other_var".
   */
  void add_mapped_symbol_to_reference(SgFunctionCallExp *func_call,
           SgFunctionDefinition *func_def);

  /**
   * \fn    set_socket_attribute_for_dmi_object(
    SgFunctionCallExp *func_call,
    SgVariableDefinition* socket)
   * \brief we need to know which module instance this dmi object connects to
   */
  void set_socket_attribute_for_dmi_object(
    SgFunctionCallExp *func_call,
    SgVariableDefinition* socket);

  /**
   * \fn    reference_analysis_hack_for_dmi_get_ptr(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def,
    SgVariableDefinition *socket)
   * \brief get_dmi_ptr is a special case. In this case, the 2nd 
   *        parameter is a reference referring to a dmi object.
   *        However, in the analysis, it should be the dmi object refering 
   *        to the parameter
   */
  void reference_analysis_hack_for_dmi_get_ptr(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def,
    SgVariableDefinition *socket);


  /**
   * \fn    reference_analysis_hack_for_b_transport(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def,
    SgVariableDefinition *socket)
   * \brief we now also keep the pcp 
   */
  void reference_analysis_hack_for_b_transport(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def,
    SgVariableDefinition *socket);

   /**
   * \fn    reference_analysis_hack_for_nb_transport(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def,
    SgVariableDefinition *socket)
   * \brief we now also keep the pcp 
   */
  void reference_analysis_hack_for_nb_transport(
    SgFunctionCallExp *func_call,
    SgFunctionDefinition *func_def,
    SgVariableDefinition *socket);


  /**
   * \fn    forward_reference_function_parameters(SgFunctionCallExp *func_call_exp)
   * \brief This function forwards the arguments which are passed by reference
   *        to the function parameters of the called function. Recursively,
   *        this function traverses through the body of the function definition
   *        and follow other function calls.
   */
  void forward_reference_function_parameters(SgFunctionCallExp *func_call_exp);

  /**
   * \fn    mark_not_conflict_free_func_calls_in_segment(
           SgNode *stmt,
           SegmentSet segments)
   * \brief This function checks if there is a not-conflict-free function call
   *        in the expression. If so, all segments marked as in conflict.
   */
  void mark_not_conflict_free_func_calls_in_segment(
           SgNode *stmt,
           SegmentSet segments);


  /**
   * \fn    analyze_prepended_pragmas_for_annotation(SgFunctionDeclaration *func_decl)
   * \brief This function checks if the function has any pragma.
   *        The results are in the return value
   */
  FunctionAnnotation*
  analyze_prepended_pragmas_for_annotation(SgFunctionDeclaration *func_decl);

  /**
   * \fn    duplicate_segments(SegmentSet segments)
   * \brief This function dupilcates a SegmentSet.
   *        the dulicated segments will have also
   *        incoming edges
   */
  SegmentSet duplicate_segments(SegmentSet segments);

  /**
   * \fn    duplicate_empty_segments(SegmentSet segments)
   * \brief This function dupilcates a SegmentSet.
   *        The dulicated segments will have also
   *        incoming edges.
   *        Duplicated segments are not inlined_segment
   */
  SegmentSet duplicate_empty_segments(SegmentSet segments);

  /**
   * \brief Contains all boudaries (currently, wait statemens)
   */
  FunctionCallBoundarySet function_boundaries_;

  /**
   * \brief Currently not used
   */
  CCxxKeywordsBoundarySet keyword_boundaries_;

  /**
   * \brief Contains the entrance vertexes of recursive function calls 
   */
  std::set<VertexDescriptor> recursive_function_calls_;

  /**
   * \brief This counter is used for generating variables names with unique
   *        names
   */
  static int counter;

  /**
   * \fn    follow_function_call(SgFunctionDeclaration *func_decl)
   * \brief true, if the function be followed.
   */
  bool follow_function_call(SgFunctionDeclaration *func_decl);

  /**
   * \fn    set_socket_call_CachedFunctionAstAttributes(
    SgFunctionDeclaration *func_decl,
    bool duplicate_segments,
    PortCallPath pcp
  )
   * \brief set the CachedFunctionAstAttributes for registered b_transport functions
   */
  void set_socket_call_CachedFunctionAstAttributes(
    SgFunctionDeclaration *func_decl,
    bool duplicate_segments,
    PortCallPath pcp
  );


  
public:
  /**
  * \fn    analyze_TLM2_container_final_mapping()
  * \brief there may be nested b_transports, so, a reference typed gp (gp0) may refer
  *        to another reference typed gp (gp1), and gp1 refers to a real gp (gp2).
  *        The function would set gp0 referring to gp2
  */
  void analyze_TLM2_container_final_mapping();
  
  /**
   * \fn    has_transition(int seg_id_from, int seg_id_to)
   * \brief true, if there is a transition from 'from' to 'to'
   */
  bool has_transition(int seg_id_from, int seg_id_to);

  /**
   * \fn    id_to_vertex_descriptor(int seg_id)
   * \brief Returns a vertex descriptor of the segment with the id 'seg_id'
   */
  VertexDescriptor id_to_vertex_descriptor(int seg_id);

  /**
   * \brief Internal data structure of the segment graph
   */
  Graph graph_;

  /**
   * \brief This set stores segments from precached functions.
   *        These segments will be deleted later.
   */
  std::set<VertexDescriptor> precached_function_segments_;

  /**
   * \brief This set stores all function calls which belong to a port.
   *        These segments will be deleted later
   */
  std::set<SgFunctionCallExp*> port_calls_;


  /**
   * \brief Is true if the segment graph is successfully created
   */
  bool is_succesfully_created_;

  /**
   * \brief used for instrumenting seg id offset in b_transport
            implementation functions.
   */
  std::unordered_map<SgFunctionDeclaration*, 
                     std::vector<int> >
                     func_socket_offset_mapping_;

  /**
   * \brief used for instrumenting seg id offset in b_transport
            implementation functions.
   */
  std::unordered_map<SgFunctionCallExp*, //socket_x->b_transport(..)
                     std::vector<int> > // <socket id, offset to be instrumented before the bt>
                     bt_socket_offset_mapping_;

  std::unordered_set<SgFunctionCallExp*> b_transport_calls_;

  /**
   * \brief including generic payload and dmi
  */ 
  std::unordered_set<SgVariableSymbol*> Tlm2_containers_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* SEGMENT_GRAPH_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
