// delta_pipe3x3.cc: delta-cycle pipeline with 3x3 inner stages
// 02/13/20, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define LANES      3
#define STAGES     3
#define ITEMS      9
#define WORKLOAD   10000000*STAGES*LANES
//#define VERBOSE

#pragma RISC M conflict-free
int M[LANES][STAGES+1][2] = { }; // memory for double-buffering

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Source)
{
    int Buffer, Index, Count;
    sc_event &OutDataReadyA, &OutBufferClearA;
    sc_event &OutDataReadyB, &OutBufferClearB;
    sc_event &OutDataReadyC, &OutBufferClearC;

    void main(void)
    {
	// send data item 1
	Count++;
	for(int l=0; l<LANES; l++)
	{
	    M[l][Buffer][Index] = (l+1)*10+Count;
	}
#ifdef VERBOSE
	printf("%2u: %s sends data %d\n", (unsigned)sc_delta_count(), name(), Count);
#endif
	OutDataReadyA.notify(SC_ZERO_TIME);
	OutDataReadyB.notify(SC_ZERO_TIME);
	OutDataReadyC.notify(SC_ZERO_TIME);
	wait(SC_ZERO_TIME);

	// send data item 2, 3, 4, ...
	for(int i=1; i<ITEMS; i++)
	{
	    Count++;
	    Index ^= 1;
	    for(int l=0; l<LANES; l++)
	    {
		M[l][Buffer][Index] = (l+1)*10+Count;
	    }
#ifdef VERBOSE
	    printf("%2u: %s sends data %d\n", (unsigned)sc_delta_count(), name(), Count);
#endif
	    OutDataReadyA.notify(SC_ZERO_TIME);
	    OutDataReadyB.notify(SC_ZERO_TIME);
	    OutDataReadyC.notify(SC_ZERO_TIME);
	    wait(OutBufferClearA & OutBufferClearB & OutBufferClearC);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp() == SC_ZERO_TIME);
	assert(sc_delta_count() == ITEMS);
    }

    Source(sc_module_name _nm, int _b,
		sc_event &_drA, sc_event &_bcA,
		sc_event &_drB, sc_event &_bcB,
		sc_event &_drC, sc_event &_bcC)
    :	sc_module(_nm), Buffer(_b),
	OutDataReadyA(_drA), OutBufferClearA(_bcA),
	OutDataReadyB(_drB), OutBufferClearB(_bcB),
	OutDataReadyC(_drC), OutBufferClearC(_bcC)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Source);
	SC_THREAD(main);
    }
};

SC_MODULE(Stage)
{
    int Lane, InBuffer, OutBuffer, Index, d;
    sc_event &InDataReady, &InBufferClear;
    sc_event &OutDataReady, &OutBufferClear;
    unsigned long w;

    void main(void)
    {
	// wait for data
	wait(InDataReady);

	// process data item 1
	d = M[Lane][InBuffer][Index];
	w = Work(WORKLOAD);
	M[Lane][OutBuffer][Index] = d;
#ifdef VERBOSE
	printf("%2u: %s processes data %d\n", (unsigned)sc_delta_count(), name(), d);
#endif
	InBufferClear.notify(SC_ZERO_TIME);
	OutDataReady.notify(SC_ZERO_TIME);
	wait(InDataReady);

	// process data item 2, 3, 4, ...
	for(;;)
	{
	    d = M[Lane][InBuffer][Index^=1];
	    w = Work(WORKLOAD);
	    M[Lane][OutBuffer][Index] = d;
#ifdef VERBOSE
	    printf("%2u: %s processes data %d\n", (unsigned)sc_delta_count(), name(), d);
#endif
	    InBufferClear.notify(SC_ZERO_TIME);
	    OutDataReady.notify(SC_ZERO_TIME);
	    wait(InDataReady & OutBufferClear);
	}
    }

    Stage(sc_module_name _nm, int _l, int _ib, int _ob,
		sc_event &_idr, sc_event &_ibc,
		sc_event &_odr, sc_event &_obc)
    :	sc_module(_nm), Lane(_l), InBuffer(_ib), OutBuffer(_ob),
	InDataReady(_idr), InBufferClear(_ibc),
	OutDataReady(_odr), OutBufferClear(_obc)
    {
	Index = 0;
	SC_HAS_PROCESS(Stage);
	SC_THREAD(main);
    }
};

SC_MODULE(Sink)
{
    int Buffer, Index, d[LANES], Count;
    sc_event &InDataReadyA, &InBufferClearA;
    sc_event &InDataReadyB, &InBufferClearB;
    sc_event &InDataReadyC, &InBufferClearC;

    void main(void)
    {
	// receive data items 1, 2, 3, ...
	for(int i=0; i<ITEMS; i++)
	{
	    wait(InDataReadyA & InDataReadyB & InDataReadyC);
	    for(int l=0; l<LANES; l++)
	    {
		d[l] = M[l][Buffer][Index];
	    }
#ifdef VERBOSE
	    printf("%2u: %s received data %d\n", (unsigned)sc_delta_count(), name(), d[0]%10);
#endif
	    InBufferClearA.notify(SC_ZERO_TIME);
	    InBufferClearB.notify(SC_ZERO_TIME);
	    InBufferClearC.notify(SC_ZERO_TIME);
	    Index^=1;
	    Count++;
	    for(int l=0; l<LANES; l++)
	    {
		assert(d[l] == (l+1)*10+Count);
	    }
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp() == SC_ZERO_TIME);
	assert(sc_delta_count() == ITEMS+STAGES);
    }

    Sink(sc_module_name _nm, int _b,
		sc_event &_drA, sc_event &_bcA,
		sc_event &_drB, sc_event &_bcB,
		sc_event &_drC, sc_event &_bcC)
    :	sc_module(_nm), Buffer(_b),
	InDataReadyA(_drA), InBufferClearA(_bcA),
	InDataReadyB(_drB), InBufferClearB(_bcB),
	InDataReadyC(_drC), InBufferClearC(_bcC)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Sink);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    sc_event data_readyA1, buffer_clearA1;
    sc_event data_readyB1, buffer_clearB1;
    sc_event data_readyC1, buffer_clearC1;
    sc_event data_readyA2, buffer_clearA2;
    sc_event data_readyB2, buffer_clearB2;
    sc_event data_readyC2, buffer_clearC2;
    sc_event data_readyA3, buffer_clearA3;
    sc_event data_readyB3, buffer_clearB3;
    sc_event data_readyC3, buffer_clearC3;
    sc_event data_readyA4, buffer_clearA4;
    sc_event data_readyB4, buffer_clearB4;
    sc_event data_readyC4, buffer_clearC4;

    Source src;
    Stage  sA1, sB1, sC1;
    Stage  sA2, sB2, sC2;
    Stage  sA3, sB3, sC3;
    Sink   snk;

    SC_CTOR(TOP)
    : src("src", 0, data_readyA1, buffer_clearA1,
			data_readyB1, buffer_clearB1,
			data_readyC1, buffer_clearC1),
	sA1("sA1", 0, 0, 1, data_readyA1, buffer_clearA1, data_readyA2, buffer_clearA2),
	sB1("sB1", 1, 0, 1, data_readyB1, buffer_clearB1, data_readyB2, buffer_clearB2),
	sC1("sC1", 2, 0, 1, data_readyC1, buffer_clearC1, data_readyC2, buffer_clearC2),
	sA2("sA2", 0, 1, 2, data_readyA2, buffer_clearA2, data_readyA3, buffer_clearA3),
	sB2("sB2", 1, 1, 2, data_readyB2, buffer_clearB2, data_readyB3, buffer_clearB3),
	sC2("sC2", 2, 1, 2, data_readyC2, buffer_clearC2, data_readyC3, buffer_clearC3),
	sA3("sA3", 0, 2, 3, data_readyA3, buffer_clearA3, data_readyA4, buffer_clearA4),
	sB3("sB3", 1, 2, 3, data_readyB3, buffer_clearB3, data_readyB4, buffer_clearB4),
	sC3("sC3", 2, 2, 3, data_readyC3, buffer_clearC3, data_readyC4, buffer_clearC4),
	snk("snk", 3, data_readyA4, buffer_clearA4,
			data_readyB4, buffer_clearB4,
			data_readyC4, buffer_clearC4)
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("delta_pipe3x3 starts\n");
    sc_start();
    assert(top.src.Count == ITEMS);
    assert(top.snk.Count == ITEMS);
    assert(sc_time_stamp() == SC_ZERO_TIME);
    printf("delta_pipe3x3 done.\n");
    printf("[potential speedup over SEQ: OOO 9x, NPD 9x, SYN 9x]\n");
    return 0;
}

// EOF
