// timed_pipe3.cc: timed pipeline with 3 inner stages
// (inspired from delta_pipe3.cc)
// 05/02/20, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define BUFSIZE    2
#define STAGES     3
#define ITEMS      10
#define WORKLOAD   10000000*STAGES
#define WORKTIME   5.0
//#define VERBOSE

#pragma RISC M conflict-free
int M[STAGES+1][BUFSIZE] = { }; // memory for double-buffering

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Source)
{
    int Buffer, Index, Count;
    sc_in<unsigned>  OutNumRcvd;
    sc_out<unsigned> OutNumSent;

    void main(void)
    {
	for(int i=0; i<ITEMS; i++)
	{
	    // make an item
	    wait(WORKTIME, SC_PS);
	    Count++;
	    wait(WORKTIME, SC_PS);

	    // if buffer is full, wait for free space
	    while (OutNumSent - OutNumRcvd >= BUFSIZE)
	    {
//		printf("%2u:%02u: DEBUG Source waiting for output buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSent - OutNumRcvd);
		wait(OutNumRcvd.value_changed_event());
	    }

	    // store item in buffer and notify receiver
	    M[Buffer][Index] = Count;
	    OutNumSent = OutNumSent + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), Count);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next send
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() == (unsigned)(2*WORKTIME*ITEMS));
    }

    Source(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Source);
	SC_THREAD(main);
    }
};

SC_MODULE(Stage)
{
    int InBuffer, OutBuffer, Index, d;
    sc_in<unsigned>  InNumSent;
    sc_in<unsigned>  OutNumRcvd;
    sc_out<unsigned> InNumRcvd;
    sc_out<unsigned> OutNumSent;
    unsigned long w;

    void main(void)
    {
	for(;;)
	{
	    // if input buffer is empty, wait for items
	    while (InNumSent - InNumRcvd == 0)
	    {
//		printf("%2u:%02u: DEBUG Stage waiting for input buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSent - InNumRcvd);
		wait(InNumSent.value_changed_event());
	    }

	    // load item from buffer and notify sender
	    d = M[InBuffer][Index];
	    InNumRcvd = InNumRcvd + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next transaction

	    // process the item
#ifdef VERBOSE
	    printf("%2u:%02u: %s processes data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    wait(WORKTIME, SC_PS);
	    w = Work(WORKLOAD);
	    wait(WORKTIME, SC_PS);

	    // if output buffer is full, wait for free space
	    while (OutNumSent - OutNumRcvd >= BUFSIZE)
	    {
//		printf("%2u:%02u: DEBUG Stage waiting for output buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSent - OutNumRcvd);
		wait(OutNumRcvd.value_changed_event());
	    }

	    // store item in output buffer and notify receiver
	    M[OutBuffer][Index] = d;
	    OutNumSent = OutNumSent + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next transaction
	}
    }

    Stage(sc_module_name _nm, int _ib, int _ob)
    :	sc_module(_nm), InBuffer(_ib), OutBuffer(_ob)
    {
	Index = 0;
	SC_HAS_PROCESS(Stage);
	SC_THREAD(main);
    }
};

SC_MODULE(Sink)
{
    int Buffer, Index, d, Count;
    sc_in<unsigned>  InNumSent;
    sc_out<unsigned> InNumRcvd;

    void main(void)
    {
	for(int i=0; i<ITEMS; i++)
	{
	    // if buffer is empty, wait for items
	    while (InNumSent - InNumRcvd == 0)
	    {
//		printf("%2u:%02u: DEBUG Sink waiting for item S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSent - InNumRcvd);
		wait(InNumSent.value_changed_event());
	    }

	    // load items from buffer and notify sender
	    d = M[Buffer][Index];
	    InNumRcvd = InNumRcvd + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next receive

	    // check the item
	    wait(WORKTIME, SC_PS);
	    Count++;
	    assert(d == Count);
	    wait(WORKTIME, SC_PS);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() == (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    }

    Sink(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Sink);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    sc_signal<unsigned> NumSent1, NumRcvd1;
    sc_signal<unsigned> NumSent2, NumRcvd2;
    sc_signal<unsigned> NumSent3, NumRcvd3;
    sc_signal<unsigned> NumSent4, NumRcvd4;

    Source src;
    Stage  s1, s2, s3;
    Sink   snk;

    SC_CTOR(TOP)
    : NumSent1("NumSent1", 0),
	NumRcvd1("NumRcvd1", 0),
	NumSent2("NumSent2", 0),
	NumRcvd2("NumRcvd2", 0),
	NumSent3("NumSent3", 0),
	NumRcvd3("NumRcvd3", 0),
	NumSent4("NumSent4", 0),
	NumRcvd4("NumRcvd4", 0),
	src("src", 0),
	s1("s1", 0, 1),
	s2("s2", 1, 2),
	s3("s3", 2, 3),
	snk("snk", 3)
    {
	src.OutNumSent.bind(NumSent1);
	src.OutNumRcvd.bind(NumRcvd1);
	s1.InNumSent.bind(NumSent1);
	s1.InNumRcvd.bind(NumRcvd1);
	s1.OutNumSent.bind(NumSent2);
	s1.OutNumRcvd.bind(NumRcvd2);
	s2.InNumSent.bind(NumSent2);
	s2.InNumRcvd.bind(NumRcvd2);
	s2.OutNumSent.bind(NumSent3);
	s2.OutNumRcvd.bind(NumRcvd3);
	s3.InNumSent.bind(NumSent3);
	s3.InNumRcvd.bind(NumRcvd3);
	s3.OutNumSent.bind(NumSent4);
	s3.OutNumRcvd.bind(NumRcvd4);
	snk.InNumSent.bind(NumSent4);
	snk.InNumRcvd.bind(NumRcvd4);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("timed_pipe3 starts\n");
    sc_start();
    assert(top.src.Count == ITEMS);
    assert(top.snk.Count == ITEMS);
    assert(sc_time_stamp().value() == (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    printf("timed_pipe3 done.\n");
    printf("[potential speedup over SEQ: OOO 3x, NPD 3x, SYN 3x]\n");
    return 0;
}

// EOF
